/*****************************************************************************/
/*                                                                           */
/* liburl2cat : Domain Categorization Library                                */
/*                                                                           */
/*  Copyright (C) 2020 Kuzmin Yaroslav <kuzmin.yaroslav@gmail.com>           */
/*                                                                           */
/* liburl2cat is free software: you can redistribute it and/or modify it     */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation, either version 3 of the License, or             */
/* (at your option) any later version.                                       */
/*                                                                           */
/* liburl2cat is distributed in the hope that it will be useful, but         */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      */
/* See the GNU General Public License for more details.                      */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program.  If not, see <http://www.gnu.org/licenses/>.           */
/*                                                                           */
/*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <stdint.h>
#include <errno.h>
#include <sys/mman.h>

#include <url2cat.h>

/*****************************************************************************/
static char* programm_name;
static const char* const short_options = "s:u:d:r:h";
static const struct option long_options[] = {
	{ "setting", 1, NULL, 's' },
	{ "url", 1, NULL, 'u' },
	{ "download", 1, NULL, 'd' },
	{ "recategory", 1, NULL, 'r' },
	{ "help", 0, NULL, 'h' },
	{ NULL, 0, NULL, 0 }
};

static void print_help(void)
{
	fprintf(stdout, "Using %s option \n", programm_name);
	fprintf(stdout, "    -s  --setting     Settings file (default: url2cat.ini)\n"
	                "    -u  --url         File with URL (default: domain.cvs)\n"
	                "    -d  --download    Download database\n"
	                "    -r  --recategory  URL recategorization (example: \"sex.com : porno\")\n"
	                "    -h  --help        Help\n");
}

/*****************************************************************************/
struct verificatio
{
	char* url;
	uint8_t type[MAX_NUMBER_CATEGORY];
};
typedef struct verificatio verificatio_s;

static verificatio_s* array_begin = NULL;
static size_t array_size = 0;
static size_t array_len = 0;
#define DEFAULT_ARRAY_SIZE 256

static int malloc_array_verification(void)
{
	verificatio_s* tmp;

	if (array_size == 0) {
		array_begin = calloc(DEFAULT_ARRAY_SIZE, sizeof(verificatio_s));
		array_size = DEFAULT_ARRAY_SIZE;
		return 0;
	}
	tmp = calloc((DEFAULT_ARRAY_SIZE + array_size), sizeof(verificatio_s));
	memcpy(tmp, array_begin, (array_size * sizeof(verificatio_s)));
	free(array_begin);
	array_begin = tmp;
	array_size += DEFAULT_ARRAY_SIZE;

	return 0;
}

static int free_array_verification(void)
{
	verificatio_s* current;
	if (array_size == 0)
		return 0;

	current = array_begin;
	for (int i = 0; i < array_len; ++i) {
		if (current->url != NULL)
			free(current->url);
		++current;
	}

	free(array_begin);
	array_begin = NULL;
	array_size = 0;
	return 0;
}

static int add_array_verification(char* url, uint8_t type[MAX_NUMBER_CATEGORY])
{
	verificatio_s* current;

	if (array_len == array_size)
		malloc_array_verification();
	current = array_begin + array_len;
	current->url = url;
	memcpy(current->type, type, (MAX_NUMBER_CATEGORY * sizeof(uint8_t)));
	++array_len;
	return 0;
}

static verificatio_s* get_array_verification(void)
{
	return array_begin;
}

static size_t get_array_len_verification(void)
{
	return array_len;
}
// https://gcc.gnu.org/onlinedocs/gcc-4.8.5/cpp/Stringification.html
#define xstr(s)         str(s)
#define str(s)          #s
#define DOMAIN_STR_SIZE 256
#define PATH_STR_SIZE   256
#define CAT_STR_SIZE    32

static int fill_verification_struct(char* name_file, verificatio_s** verificatio)
{
	int rc;
	FILE* stream_file;
	char domain_str[DOMAIN_STR_SIZE + 1] = { 0 };
	char path_str[PATH_STR_SIZE + 1] = { 0 };
	char type_str[CAT_STR_SIZE + 1] = { 0 };
	size_t len;
	uint8_t type[MAX_NUMBER_CATEGORY];

	stream_file = fopen(name_file, "r");
	if (stream_file == NULL) {
		fprintf(stderr, "Can not open %s file: %s\n", name_file, strerror(errno));
		return -1;
	}

	for (;;) {
		rc = fscanf(stream_file, "%" xstr(DOMAIN_STR_SIZE) "s\t%" xstr(PATH_STR_SIZE) "s\t%" xstr(CAT_STR_SIZE) "s\n", domain_str, path_str, type_str);
		if (rc != 3) {
			break;
		}

		len = strlen(domain_str);
		len += strlen(path_str);
		char* url_str = calloc((len + 1), sizeof(char));
		strcat(url_str, domain_str);
		strcat(url_str, path_str);

		char* beginptr = type_str;
		char* endptr = NULL;
		memset(type, 0, (MAX_NUMBER_CATEGORY * sizeof(uint8_t)));
		for (int i = 0; i < MAX_NUMBER_CATEGORY; ++i) {
			type[i] = strtoll(beginptr, &endptr, 10);
			if (*endptr == 0)
				break;
			++endptr;
			beginptr = endptr;
		}

		add_array_verification(url_str, type);
	}

	*verificatio = get_array_verification();
	fclose(stream_file);
	return 0;
}

static int free_verification_struct(void)
{
	free_array_verification();
	return 0;
}

/*****************************************************************************/

static int check_type(uint8_t type, uint8_t verificatio_type[MAX_NUMBER_CATEGORY])
{
	for (int i = 0; i < MAX_NUMBER_CATEGORY; ++i) {
		if (type == verificatio_type[i]) {
			return 1;
		}
	}
	return 0;
}

/*****************************************************************************/
static char* setting_file = NULL;
static char setting_file_default[] = "url2cat.ini";
static char* verificatio_file = NULL;
static char verificatio_file_default[] = "domains.csv";
static url2cat_setting_s setting = { 0 };
static char* recategory_str = NULL;
static size_t recategory_str_size = 0;

#define YES 0
#define NO  1

int main(int argc, char* argv[])
{
	int rc;
	int next_option = 0;
	verificatio_s* verificatio;
#ifdef HAVE_DATABASE
	int flag_download = NO;
#endif
	int flag_recategory = NO;
	programm_name = argv[0];

	printf("Sample query - %s \n", url2cat_version());

	for (; next_option != -1;) {
		next_option = getopt_long(argc, argv, short_options, long_options, NULL);
		switch (next_option) {
			case 'd':
#ifdef HAVE_DATABASE
				flag_download = YES;
#else
				fprintf(stderr, "%s utility does not support working with the database\n", programm_name);
				exit(0);
#endif
				break;
			case 's':
				setting_file = strdup(optarg);
				break;
			case 'u':
				verificatio_file = strdup(optarg);
				break;
			case 'r':
				flag_recategory = YES;
				recategory_str = strdup(optarg);
				recategory_str_size = strlen(recategory_str);
				break;
			case 'h':
			case '?':
				print_help();
				exit(SUCCESS);
				break;
			case -1:
				break;
			default:
				exit(SUCCESS);
		}
	}

	if (setting_file == NULL)
		setting_file = setting_file_default;

	rc = url2cat_read_ini(setting_file, &setting);
	if (rc != SUCCESS) {
		fprintf(stderr, "Error read ini file %s : %x\n", setting_file, rc);
		return 0;
	}

	rc = url2cat_init(&setting);
	if (rc != SUCCESS) {
		fprintf(stderr, "url2cat init : error %x\n", rc);
		return 0;
	}

#ifdef HAVE_DATABASE
	if (flag_download == YES) {
		rc = url2cat_update_database(&setting);
		if (rc != SUCCESS)
			fprintf(stderr, "url2cat update database : error %x\n", rc);
		else
			fprintf(stdout, "url2cat update database : success!\n");
		return 0;
	}
#endif

	if (flag_recategory == YES) {
		fprintf(stderr, "url2cat recategory : %s\n", recategory_str);
		//char url_recat[] = "sex.com";
		//char type_recat[] = "porno";
		//rc = url2cat_recategory(url_recat , strlen(url_recat), type_recat, strlen(type_recat));
		return 0;
	}

	if (verificatio_file == NULL)
		verificatio_file = verificatio_file_default;

	rc = fill_verification_struct(verificatio_file, &verificatio);
	if (rc != 0) {
		fprintf(stderr, "Error fill verificatio struct\n");
		return 0;
	}

	verificatio_s* current = get_array_verification();
	for (int i = 0; i < get_array_len_verification(); ++i) {
		char* url = current->url;
		size_t url_len = strlen(url);
		url2cat_category_s category[MAX_NUMBER_CATEGORY];
		int category_number = MAX_NUMBER_CATEGORY;

		rc = url2cat_category(url, url_len, category, &category_number);
		printf("%04d: url %s\n", i, current->url);
		if (rc != SUCCESS) {
			printf("     categorization error : %x\n", rc);
		}
		else {
			printf("     category : ");
			for (int j = 0; j < category_number; ++j) {
				rc = check_type(category[j].type, current->type);
				if (rc) {
					printf(" %d, ", category[j].type);
				}
				else {
					printf(" %d (not found), ", category[j].type);
				}
			}
			printf("\n");
		}
		++current;
	}
	rc = url2cat_deinit();
	if (rc != SUCCESS) {
		fprintf(stderr, "url2cat deinit : error %x\n", rc);
	}

	free_verification_struct();
	return 0;
}
/*****************************************************************************/
