/*****************************************************************************/
/*                                                                           */
/* liburl2cat : Domain Categorization Library                                */
/*                                                                           */
/*  Copyright (C) 2020 Kuzmin Yaroslav <kuzmin.yaroslav@gmail.com>           */
/*  Copyright (C) 2022 Kuzmin Yaroslav <kuzmin.yaroslav@gmail.com>           */
/*                                                                           */
/* liburl2cat is free software: you can redistribute it and/or modify it     */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation, either version 3 of the License, or             */
/* (at your option) any later version.                                       */
/*                                                                           */
/* liburl2cat is distributed in the hope that it will be useful, but         */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      */
/* See the GNU General Public License for more details.                      */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program.  If not, see <http://www.gnu.org/licenses/>.           */
/*                                                                           */
/*****************************************************************************/

#include "common.h"
#include "locale.h"

/*****************************************************************************/

static uint8_t* cache_buffer = NULL;
static size_t cache_size = 0;
static pthread_rwlock_t cache_mutex;

#define SIZE_CELL (DOMAIN_HASH_SIZE + PATH_HASH_SIZE + MAX_NUMBER_CATEGORY)

#ifdef CACHE_DINAMIC

	#define SIZE_FLAG 1
	#if (SIZE_CELL > 255)
		#error "Dynamic cache does not support cell size over 255!"
	#endif
static uint8_t* cache_buffer_end = NULL;
static uint8_t* cache_end = NULL;
	#define OFFSET_MAX (SIZE_CELL + 1)

#elif CACHE_STATIC

static int cells_number_total = 0;
static int cells_number_full = 0;
	#if SIZE_CELL > 32
		#define SIZE_TO_NUMBER_CELL(size)   (size / SIZE_CELL)
		#define NUMBER_TO_SIZE_CELL(number) (number * SIZE_CELL)
	#else
		#define SIZE_CELL_ALIGNED           5
		#define SIZE_TO_NUMBER_CELL(size)   (size >> SIZE_CELL_ALIGNED)
		#define NUMBER_TO_SIZE_CELL(number) (number << SIZE_CELL_ALIGNED)
	#endif

#else

	#error "Select type cache"

#endif

#define CACHE_LIMIT_SIZE 0x2000000     // 32Mb

#ifdef CACHE_STATIC
static uint8_t* search(uint8_t* cell, int* found)
{
	if (cells_number_full == 0) {
		*found = 0;
		return NULL;
	}
	uint8_t* begin = cache_buffer;
	uint8_t* end = cache_buffer + (NUMBER_TO_SIZE_CELL(cells_number_full));
	size_t off = NUMBER_TO_SIZE_CELL((cells_number_full >> 1));
	uint8_t* cur = cache_buffer + off;

	int check = 0;
	for (;;) {
		check = memcmp(cur, cell, (DOMAIN_HASH_SIZE + PATH_HASH_SIZE));
		if (check == 0)
			break;
		if (check < 0) {
			if (begin == cur) {
				cur += NUMBER_TO_SIZE_CELL(1);
				break;
			}
			begin = cur;
		}
		else {
			if (end == cur) {
				break;
			}
			end = cur;
		}

		off = SIZE_TO_NUMBER_CELL((end - begin));
		off >>= 1;
		off = NUMBER_TO_SIZE_CELL(off);

		cur = begin + off;
	}
	if (cur == (cache_buffer + (NUMBER_TO_SIZE_CELL(cells_number_full))))
		cur -= NUMBER_TO_SIZE_CELL(1);
	*found = SIZE_TO_NUMBER_CELL((cur - cache_buffer));

	if (check != 0)
		cur = NULL;

	return cur;
}

#endif

int get_category_cache(char* url_str, size_t url_str_size, url2cat_category_s* p_category, int* p_number)
{
	if (cache_size == 0)
		return CACHE_URL_NOT_FOUND;

	char* domain = url_str;
	char* path = memchr(url_str, '/', url_str_size);
	if (path == NULL)
		return CACHE_URL_NO_PATH;
	size_t len_domain = path - url_str;
	size_t len_path = url_str_size - len_domain;

#ifdef CACHE_DINAMIC
	uint8_t hash_domain[DOMAIN_HASH_SIZE];
	uint8_t* hash = COMPUTES_HASH((const uint8_t*)domain, len_domain, NULL);
	memmove(hash_domain, hash, DOMAIN_HASH_SIZE);
	uint8_t hash_path[PATH_HASH_SIZE] = { 0 };
	if (len_path > sizeof(char)) {
		hash = COMPUTES_HASH((const uint8_t*)path, len_path, NULL);
		memmove(hash_path, hash, PATH_HASH_SIZE);
	}

	pthread_rwlock_rdlock(&cache_mutex);
	uint8_t* cache = cache_buffer;
	uint8_t flag;
	int found = 0;

	for (; cache < cache_end;) {
		flag = *cache;
		if ((flag > (SIZE_FLAG + DOMAIN_HASH_SIZE + MAX_NUMBER_CATEGORY)) && (hash_path[0] == 0)) {
			cache += flag;
			continue;
		}
		int check = memcmp(hash_domain, (cache + SIZE_FLAG), DOMAIN_HASH_SIZE);
		if (check == 0) {
			if (flag <= (SIZE_FLAG + DOMAIN_HASH_SIZE + MAX_NUMBER_CATEGORY)) {
				++found;
				break;
			}
			else {
				check = memcmp(hash_path, (cache + SIZE_FLAG + DOMAIN_HASH_SIZE), PATH_HASH_SIZE);
				if (check == 0) {
					++found;
					break;
				}
			}
		}
		cache += flag;
	}
	if (found) {
		uint8_t* type = cache;
		if (flag <= (SIZE_FLAG + DOMAIN_HASH_SIZE + MAX_NUMBER_CATEGORY)) {
			type += (SIZE_FLAG + DOMAIN_HASH_SIZE);
			flag -= (SIZE_FLAG + DOMAIN_HASH_SIZE);
		}
		else {
			type += (SIZE_FLAG + DOMAIN_HASH_SIZE + PATH_HASH_SIZE);
			flag -= (SIZE_FLAG + DOMAIN_HASH_SIZE + PATH_HASH_SIZE);
		}

		for (int i = 0; ((i < flag) && (i < *p_number)); ++i) {
			p_category->type = *type;
			const char* type_name = get_category_name(*type);
			if (type_name != NULL)
				strncpy(p_category->type_name, type_name, CATEGORY_TYPE_NAME_STR_SIZE);
			else
				*p_category->type_name = 0;
			++p_category;
			++type;
		}
		if (*p_number > flag)
			*p_number = flag;
		pthread_rwlock_unlock(&cache_mutex);
		return SUCCESS;
	}

#elif CACHE_STATIC
	uint8_t cell_search[SIZE_CELL] = { 0 };
	uint8_t* hash = COMPUTES_HASH((const uint8_t*)domain, len_domain, NULL);
	memmove(cell_search, hash, DOMAIN_HASH_SIZE);

	if (len_path > sizeof(char)) {
		hash = COMPUTES_HASH((const uint8_t*)path, len_path, NULL);
		memmove((cell_search + DOMAIN_HASH_SIZE), hash, PATH_HASH_SIZE);
	}

	pthread_rwlock_rdlock(&cache_mutex);

	int cell_off;
	uint8_t* cell_found = search(cell_search, &cell_off);

	if (cell_found != NULL) {
		uint8_t* type = cell_found + (DOMAIN_HASH_SIZE + PATH_HASH_SIZE);
		for (int i = 0; (i < *p_number); ++i) {
			if (*type == 0) {
				*p_number = i;
				break;
			}
			p_category->type = *type;
			const char* type_name = get_category_name(*type);
			if (type_name != NULL)
				strncpy(p_category->type_name, type_name, CATEGORY_TYPE_NAME_STR_SIZE);
			else
				*p_category->type_name = 0;
			++p_category;
			++type;
		}
		pthread_rwlock_unlock(&cache_mutex);
		return SUCCESS;
	}
#endif

	pthread_rwlock_unlock(&cache_mutex);
	return CACHE_URL_NOT_FOUND;
}

#ifdef CACHE_DINAMIC
static int cache_offset(size_t len_end)
{
	uint8_t* begin = cache_buffer;
	uint8_t* end = cache_buffer;
	size_t len_begin = 0;
	uint8_t flag = *end;
	len_begin = flag;
	for (; end < cache_buffer_end;) {
		if (len_begin > OFFSET_MAX) {
			end += flag;
			break;
		}
		end += flag;
		flag = *end;
		len_begin += flag;
	}
	memmove(begin, end, (cache_size - len_begin - len_end));
	cache_end -= len_begin;
	return SUCCESS;
}
static int cache_check_add(void)
{
	size_t len_end = cache_buffer_end - cache_end;
	if (len_end < OFFSET_MAX) {
		cache_offset(len_end);
	}
	return SUCCESS;
}
#endif

int add_category_cache(char* url_str, size_t url_str_size, url2cat_category_s* p_category, int number)
{
	if (cache_size == 0)
		return SUCCESS;
	if (number == 0)
		return SUCCESS;
	char* domain = url_str;
	char* path = memchr(url_str, '/', url_str_size);
	if (path == NULL)
		return CACHE_URL_NO_PATH;

	size_t len_domain = path - url_str;
	size_t len_path = url_str_size - len_domain;

	pthread_rwlock_wrlock(&cache_mutex);

#ifdef CACHE_DINAMIC
	cache_check_add();
	uint8_t* hash = COMPUTES_HASH((const uint8_t*)domain, len_domain, NULL);
	uint8_t* begin = cache_end;
	uint8_t* end = begin + SIZE_FLAG;
	uint8_t flag = SIZE_FLAG;
	memmove(end, hash, DOMAIN_HASH_SIZE);
	flag += DOMAIN_HASH_SIZE;
	end += DOMAIN_HASH_SIZE;
	if (len_path > sizeof(char)) {
		hash = COMPUTES_HASH((const uint8_t*)path, len_path, NULL);
		memmove(end, hash, PATH_HASH_SIZE);
		flag += PATH_HASH_SIZE;
		end += PATH_HASH_SIZE;
	}
	if (number > MAX_NUMBER_CATEGORY)
		number = MAX_NUMBER_CATEGORY;
	for (int i = 0; i < number; ++i) {
		*end = p_category->type;
		++flag;
		++end;
		++p_category;
	}

	*begin = flag;
	cache_end += flag;

#elif CACHE_STATIC
	uint8_t cell_add[SIZE_CELL] = { 0 };
	uint8_t* hash = COMPUTES_HASH((const uint8_t*)domain, len_domain, NULL);
	memmove(cell_add, hash, DOMAIN_HASH_SIZE);

	if (len_path > sizeof(char)) {
		hash = COMPUTES_HASH((const uint8_t*)path, len_path, NULL);
		memmove((cell_add + DOMAIN_HASH_SIZE), hash, PATH_HASH_SIZE);
	}

	if (number > MAX_NUMBER_CATEGORY)
		number = MAX_NUMBER_CATEGORY;
	uint8_t* type = cell_add + DOMAIN_HASH_SIZE + PATH_HASH_SIZE;
	for (int i = 0; i < number; ++i) {
		*type = p_category->type;
		++type;
		++p_category;
	}

	int cell_off;
	uint8_t* cell_found = search(cell_add, &cell_off);
	if (cell_found == NULL) {
		uint8_t* src = cache_buffer + (NUMBER_TO_SIZE_CELL(cell_off));
		uint8_t* dest = src + NUMBER_TO_SIZE_CELL(1);
		if (cells_number_full != cells_number_total) {
			++cells_number_full;
		}
		int size = cells_number_full - cell_off - 1;
		memmove(dest, src, NUMBER_TO_SIZE_CELL(size));
		memmove(src, cell_add, SIZE_CELL);
	}

#endif

	pthread_rwlock_unlock(&cache_mutex);

	return SUCCESS;
}

int reset_cache(void)
{
#ifdef CACHE_DINAMIC
	cache_end = cache_buffer;
#elif CACHE_STATIC
	cells_number_full = 0;
#endif
	return SUCCESS;
}

int init_cache(size_t size)
{
	if (size > CACHE_LIMIT_SIZE)
		return CACHE_SIZE_LIMITE;
	cache_size = size;
	if (cache_size == 0)
		return SUCCESS;
	cache_buffer = calloc(sizeof(uint8_t), size);
	if (cache_buffer == NULL)
		return CACHE_BUFFER_MALLOC_NULL;
#ifdef CACHE_DINAMIC
	cache_buffer_end = cache_buffer + cache_size;
	cache_end = cache_buffer;
#elif CACHE_STATIC
	cells_number_total = SIZE_TO_NUMBER_CELL(size);
	cells_number_full = 0;
#endif
	pthread_rwlock_init(&cache_mutex, NULL);
	return SUCCESS;
}

int deinit_cache(void)
{
	if (cache_size == 0)
		return SUCCESS;
	if (cache_buffer != NULL) {
		free(cache_buffer);
		cache_buffer = NULL;
		cache_size = 0;
	}
	pthread_rwlock_destroy(&cache_mutex);
	return SUCCESS;
}
/******************************************************************************/
