/*****************************************************************************/
/*                                                                           */
/* liburl2cat : Domain Categorization Library                                */
/*                                                                           */
/*  Copyright (C) 2020 Kuzmin Yaroslav <kuzmin.yaroslav@gmail.com>           */
/*  Copyright (C) 2022 Kuzmin Yaroslav <kuzmin.yaroslav@gmail.com>           */
/*                                                                           */
/* liburl2cat is free software: you can redistribute it and/or modify it     */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation, either version 3 of the License, or             */
/* (at your option) any later version.                                       */
/*                                                                           */
/* liburl2cat is distributed in the hope that it will be useful, but         */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      */
/* See the GNU General Public License for more details.                      */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program.  If not, see <http://www.gnu.org/licenses/>.           */
/*                                                                           */
/*****************************************************************************/

#include "common.h"

/*****************************************************************************/

struct canon_buffer
{
	char* begin;
	char* end;
	size_t size;
};
typedef struct canon_buffer canon_buffer_s;

/*****************************************************************************/
static int malloc_canon_buffer(canon_buffer_s* canon_buffer, size_t size)
{
	canon_buffer->begin = malloc(size + 1);
	if (canon_buffer->begin == NULL)
		return CANON_BUFFER_MALLOC_NULL;
	canon_buffer->size = size;
	*(canon_buffer->begin + canon_buffer->size) = 0;
	return SUCCESS;
}

static int realloc_canon_buffer(canon_buffer_s* canon_buffer, size_t add_len)
{
	size_t len = canon_buffer->end - canon_buffer->begin + 1;
	if ((len + add_len) < canon_buffer->size)
		return SUCCESS;
	char* buf = malloc((len + add_len));
	if (buf == NULL)
		return CANON_BUFFER_REALLOC_NULL;
	memmove(buf, canon_buffer->begin, len);
	free(canon_buffer->begin);
	canon_buffer->begin = buf;
	canon_buffer->end = canon_buffer->begin + len - 1;
	canon_buffer->size = len + add_len;
	return SUCCESS;
}

/*****************************************************************************/
static int remove_control_character(canon_buffer_s* canon_buffer, char* url_str, size_t url_str_size)
{
	int i = 0;
	char* in = url_str;
	for (; i < url_str_size; ++i, ++in) {
		if (*in != ' ')
			break;
	}
	char* out = canon_buffer->begin;
	for (; i < url_str_size; ++i, ++in) {
		if ((*in == '\t') || (*in == '\n') || (*in == '\r') || (*in == 0)) {
			continue;
		}
		*out = *in;
		++out;
	}
	for (;;) {
		--out;
		if (*out != ' ')
			break;
	}
	++out;
	canon_buffer->end = out;
	*canon_buffer->end = 0;
	return SUCCESS;
}

static int remove_fragment(canon_buffer_s* canon_buffer)
{
	/*URL fragment              # */
	char* str = strrchr(canon_buffer->begin, '#');
	if (str != NULL) {
		*str = 0;
		canon_buffer->end = str;
	}
	return SUCCESS;
}

static char UNESCAPE_PERCENT = '%';

static int escape_hex(canon_buffer_s* canon_buffer)
{
	int rc;
	char* buf;

	buf = canon_buffer->begin;
	for (; buf < canon_buffer->end; ++buf) {
		if ((*buf <= 0x20) || (*buf >= 0xF7)) {
			size_t off = buf - canon_buffer->begin;
			rc = realloc_canon_buffer(canon_buffer, 2);
			if (rc == CANON_BUFFER_REALLOC_NULL)
				return rc;
			buf = canon_buffer->begin + off;
			char* begin = canon_buffer->end;
			char* end = canon_buffer->end + 2;
			for (; begin != buf;) {
				*end = *begin;
				--end;
				--begin;
			}
			char temp = *buf;
			*buf = UNESCAPE_PERCENT;
			++buf;
			*buf = up_hex_char(temp);
			++buf;
			*buf = low_hex_char(temp);
			canon_buffer->end += 2;
			*canon_buffer->end = 0;
		}
	}
	return SUCCESS;
}

static int char_hex(char up, char low, uint8_t* hex)
{
	uint8_t t;
	if ((up >= '0') && (up <= '9')) {
		t = up - '0';
	}
	else {
		if ((up >= 'A') && (up <= 'F')) {
			t = up - 0x37;
		}
		else {
			if ((up >= 'a') && (up <= 'f')) {
				t = up - 0x57;
			}
			else {
				return CANON_CHAR_HEX_UP;
			}
		}
	}
	*hex = t << 4;
	if ((low >= '0') && (low <= '9')) {
		t = low - '0';
	}
	else {
		if ((low >= 'A') && (low <= 'F')) {
			t = low - 0x37;
		}
		else {
			if ((low >= 'a') && (low <= 'f')) {
				t = low - 0x57;
			}
			else {
				return CANON_CHAR_HEX_LOW;
			}
		}
	}
	*hex += (t & 0x0F);
	return SUCCESS;
}

static int remove_percent_unescape(canon_buffer_s* canon_buffer)
{
	char* buf;
	char* begin;
	char* end;
	int rc;

	buf = canon_buffer->begin;
	for (; buf < canon_buffer->end;) {
		/*                       */
		if (*buf != UNESCAPE_PERCENT) {
			++buf;
			continue;
		}
		/*                %      %            %                 */
		if ((*(buf + 1) == UNESCAPE_PERCENT) || (*(buf + 1) == 0)) {
			size_t off = buf - canon_buffer->begin;
			rc = realloc_canon_buffer(canon_buffer, 2);
			if (rc == CANON_BUFFER_REALLOC_NULL)
				return rc;
			buf = canon_buffer->begin + off;
			begin = canon_buffer->end - 1;
			end = canon_buffer->end + 1;
			for (; buf < begin;) {
				*end = *begin;
				--end;
				--begin;
			}
			++buf;
			*buf = (UNESCAPE_PERCENT >> 4) + '0';
			++buf;
			*buf = (UNESCAPE_PERCENT & 0x0F) + '0';
			++buf;
			canon_buffer->end += 2;
			*canon_buffer->end = 0;
			continue;
		}

		uint8_t hex;
		rc = char_hex(*(buf + 1), *(buf + 2), &hex);
		/*                   %                      */
		if (rc != SUCCESS)
			return rc;
		/*      %                                             */
		if (((hex <= 0x20) || (hex >= 0x7F)) || (hex == 0x23)) {
			buf += 3;
			continue;
		}
		/*      %                                                 */
		if (hex != UNESCAPE_PERCENT) {
			*buf = hex;
			++buf;
			begin = buf;
			end = buf + 2;
			for (; end < canon_buffer->end;) {
				*begin = *end;
				++begin;
				++end;
			}
			canon_buffer->end -= 2;
			*canon_buffer->end = 0;
			continue;
		}
		/*                        %25 */
		hex = *(buf + 3);
		if (hex != UNESCAPE_PERCENT) {
			rc = char_hex(*(buf + 3), *(buf + 4), &hex);
			/*      %25                         */
			if (rc != SUCCESS) {
				buf += 3;
				continue;
			}
			/*       %25                                    % */
			begin = buf + 1;
			end = buf + 3;
			for (; end < canon_buffer->end;) {
				*begin = *end;
				++begin;
				++end;
			}
			canon_buffer->end -= 2;
			*canon_buffer->end = 0;
			continue;
		}
		/*       %25      %XX*/
		rc = char_hex(*(buf + 4), *(buf + 5), &hex);
		/*                        */
		if (rc != SUCCESS)
			return rc;
		/*       %25                                     6         */
		if (((hex <= 0x20) || (hex >= 0x7F)) || (hex == 0x23)) {
			buf += 6;
			continue;
		}
		/*                       %25%xx%xx */
		uint8_t hex_next;
		rc = char_hex(*(buf + 7), *(buf + 8), &hex_next);
		/*                                               */
		if (rc != SUCCESS)
			return rc;
		/*       %25                                                     */
		*(buf + 1) = hex;
		*(buf + 2) = hex_next;
		begin = buf + 3;
		end = buf + 9;
		for (; end < canon_buffer->end;) {
			*begin = *end;
			++begin;
			++end;
		}
		canon_buffer->end -= 6;
		*canon_buffer->end = 0;
		buf += 3;
	}

	return SUCCESS;
}

static char SCHEME_END[] = "://";
static int remove_scheme(canon_buffer_s* canon_buffer)
{
	char* buf = canon_buffer->begin;
	char* str = strstr(buf, SCHEME_END);
	if (str == NULL)
		return SUCCESS;
	str += 3;
	size_t len = str - buf;
	for (; str < canon_buffer->end;) {
		*buf = *str;
		++buf;
		++str;
	}
	canon_buffer->end -= len;
	*canon_buffer->end = 0;
	return SUCCESS;
}

static int remove_user_password(canon_buffer_s* canon_buffer)
{
	char* buf = canon_buffer->begin;
	char* str = buf;
	for (; str < canon_buffer->end; ++str) {
		if ((*str == '/') || (*str == '@'))
			break;
	}
	/*                  */
	if ((*str == 0) || (*str == '/'))
		return SUCCESS;
	++str;
	size_t len = str - buf;
	for (; str < canon_buffer->end;) {
		*buf = *str;
		++buf;
		++str;
	}
	canon_buffer->end -= len;
	*canon_buffer->end = 0;
	return SUCCESS;
}

static int remove_dot_host(canon_buffer_s* canon_buffer)
{
	char* buf = canon_buffer->begin;
	size_t len = 0;
	char* begin;
	char* end;
	/*                        */
	for (; buf < canon_buffer->end; ++buf) {
		if (*buf != '.')
			break;
	}
	len = buf - canon_buffer->begin;
	if (len != 0) {
		begin = canon_buffer->begin;
		end = buf;
		for (; end < canon_buffer->end;) {
			*begin = *end;
			++begin;
			++end;
		}
		canon_buffer->end -= len;
		*canon_buffer->end = 0;
	}
	/*                      */
	end = strchr(buf, '/');
	if (end == NULL) {
		end = strchr(buf, '?');
		if (end == NULL)
			end = canon_buffer->end;
	}
	--end;
	begin = end;
	for (; canon_buffer->begin < begin; --begin) {
		if (*begin != '.')
			break;
	}
	len = end - begin;
	if (len != 0) {
		++begin;
		++end;
		for (; end < canon_buffer->end;) {
			*begin = *end;
			++begin;
			++end;
		}
		canon_buffer->end -= len;
		*canon_buffer->end = 0;
	}
	/*                     */
	buf = canon_buffer->begin;
	for (; buf < canon_buffer->end;) {
		if (*buf == '/')
			break;
		if (*buf != '.') {
			++buf;
			continue;
		}
		if (*(buf + 1) != '.') {
			buf += 2;
		}
		else {
			begin = buf + 1;
			end = buf + 2;
			for (; end < canon_buffer->end;) {
				*begin = *end;
				++begin;
				++end;
			}
			--canon_buffer->end;
			*canon_buffer->end = 0;
		}
	}
	return SUCCESS;
}

static int check_uppercase_hex(char* up, char* low)
{
	if ((*up >= 'a') && (*up <= 'f'))
		*up -= 0x20;
	if ((*low >= 'a') && (*low <= 'f'))
		*low -= 0x20;
	return SUCCESS;
}

static int check_lowercase_hex(char* symbol)
{
	if ((*symbol >= 'A') && (*symbol <= 'Z'))
		*symbol += 0x20;
	return SUCCESS;
}

static int check_lowercase_host(canon_buffer_s* canon_buffer)
{
	char* buf = canon_buffer->begin;
	for (; buf < canon_buffer->end;) {
		if (*buf == '/')
			break;
		if (*buf == UNESCAPE_PERCENT) {
			check_uppercase_hex((buf + 1), (buf + 2));
			buf += 3;
		}
		check_lowercase_hex(buf);
		++buf;
	}
	return SUCCESS;
}

static int remove_port(canon_buffer_s* canon_buffer)
{
	char* begin = strchr(canon_buffer->begin, ':');
	if (begin == NULL)
		return SUCCESS;
	char* end = strchr(canon_buffer->begin, '/');
	if (end == NULL) {
		// TODO maybe  [?query] component without [/path] component
		canon_buffer->end = begin;
		*canon_buffer->end = 0;
	}
	/*         /       : */
	if (begin > end)
		return SUCCESS;
	size_t len = end - begin;
	for (; end < canon_buffer->end;) {
		*begin = *end;
		++begin;
		++end;
	}
	canon_buffer->end -= len;
	*canon_buffer->end = 0;
	return SUCCESS;
}

static int check_slash_path(canon_buffer_s* canon_buffer)
{
	int rc;
	char* path;
	char* buf;
	char* begin;
	char* end;

	path = strchr(canon_buffer->begin, '/');
	if (path == NULL) {
		rc = realloc_canon_buffer(canon_buffer, 2);
		if (rc == CANON_BUFFER_REALLOC_NULL)
			return rc;
		/*           / */
		path = strchr(canon_buffer->begin, '?');
		if (path == NULL) {
			*canon_buffer->end = '/';
			++canon_buffer->end;
			*canon_buffer->end = 0;
			return SUCCESS;
		}
		begin = canon_buffer->end - 1;
		end = canon_buffer->end;
		--path;
		for (; path < begin;) {
			*end = *begin;
			--begin;
			--end;
		}
		++path;
		*path = '/';
		++canon_buffer->end;
		*canon_buffer->end = 0;
		return SUCCESS;
	}

	buf = path;
	/*                        */
	for (; buf < canon_buffer->end;) {
		if (*buf == '?')
			break;
		if (*buf != '/') {
			++buf;
			continue;
		}

		if (*(buf + 1) != '/') {
			buf += 2;
		}
		else {
			begin = buf + 1;
			end = buf + 2;
			for (; end < canon_buffer->end;) {
				*begin = *end;
				++begin;
				++end;
			}
			--canon_buffer->end;
			*canon_buffer->end = 0;
		}
	}
	/*          /./ */
	buf = path;
	for (; buf < canon_buffer->end;) {
		if (*buf == '?')
			break;
		if (*buf != '/') {
			++buf;
			continue;
		}
		if (*(buf + 1) != '.') {
			buf += 2;
			continue;
		}
		if (*(buf + 2) != '/') {
			buf += 3;
			continue;
		}
		begin = buf + 1;
		end = buf + 3;
		for (; end < canon_buffer->end;) {
			*begin = *end;
			++begin;
			++end;
		}
		canon_buffer->end -= 2;
		*canon_buffer->end = 0;
	}
	/*            /../ */
	buf = path;
	for (; buf < canon_buffer->end;) {
		if (*buf == '?')
			break;
		if (*buf != '/') {
			++buf;
			continue;
		}
		if ((*(buf + 1) != '.') || (*(buf + 2) != '.') || (*(buf + 3) != '/')) {
			++buf;
			continue;
		}
		/* /../                         */
		if (buf == path)
			return CANON_PATH_SLASH_DOT_DOT;

		end = buf + 4;
		for (; path < buf; --buf) {
			if (*buf == '/')
				break;
		}
		begin = buf + 1;
		size_t len = end - begin;
		for (; end < canon_buffer->end;) {
			*begin = *end;
			++begin;
			++end;
		}
		canon_buffer->end -= len;
		*canon_buffer->end = 0;
	}

	return SUCCESS;
}

int get_canonicalize_url(char* url_str, size_t url_str_size, char** url_canonical_str, size_t* url_canonical_str_size)
{
	canon_buffer_s canon_buffer;

	*url_canonical_str = NULL;
	*url_canonical_str_size = 0;

	int rc = malloc_canon_buffer(&canon_buffer, url_str_size);
	if (rc != SUCCESS)
		return rc;

	*url_canonical_str = canon_buffer.begin;

	rc = remove_control_character(&canon_buffer, url_str, url_str_size);
	if (rc != SUCCESS)
		return rc;
	rc = remove_fragment(&canon_buffer);
	if (rc != SUCCESS)
		return rc;
	// TODO  check memory leak
	rc = escape_hex(&canon_buffer);
	if (rc != SUCCESS)
		return rc;
	// TODO  check memory leak
	rc = remove_percent_unescape(&canon_buffer);
	if (rc != SUCCESS)
		return rc;
	rc = remove_scheme(&canon_buffer);
	if (rc != SUCCESS)
		return rc;
	rc = remove_user_password(&canon_buffer);
	if (rc != SUCCESS)
		return rc;
	rc = remove_dot_host(&canon_buffer);
	if (rc != SUCCESS)
		return rc;
	rc = check_lowercase_host(&canon_buffer);
	if (rc != SUCCESS)
		return rc;
	rc = remove_port(&canon_buffer);
	if (rc != SUCCESS)
		return rc;
	// TODO  check memory leak
	rc = check_slash_path(&canon_buffer);
	if (rc != SUCCESS)
		return rc;

	*url_canonical_str = canon_buffer.begin;
	*url_canonical_str_size = canon_buffer.end - canon_buffer.begin;

	return SUCCESS;
}

int deinit_canonicalize_url(void)
{
	return SUCCESS;
}

/*****************************************************************************/
