/*****************************************************************************/
/*                                                                           */
/* liburl2cat : Domain Categorization Library                                */
/*                                                                           */
/*  Copyright (C) 2020 Kuzmin Yaroslav <kuzmin.yaroslav@gmail.com>           */
/*  Copyright (C) 2022 Kuzmin Yaroslav <kuzmin.yaroslav@gmail.com>           */
/*                                                                           */
/* liburl2cat is free software: you can redistribute it and/or modify it     */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation, either version 3 of the License, or             */
/* (at your option) any later version.                                       */
/*                                                                           */
/* liburl2cat is distributed in the hope that it will be useful, but         */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      */
/* See the GNU General Public License for more details.                      */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program.  If not, see <http://www.gnu.org/licenses/>.           */
/*                                                                           */
/*****************************************************************************/

#include "common.h"
#include "locale.h"
#include "net.h"
#include "sqlite3.h"

#define SHOW_DOWNLOAD 0

/*****************************************************************************/

static char database_name[DATABASE_NAME_STR_SIZE] = { 0 };
static sqlite3* database = NULL;

static pthread_mutex_t update_db_mutex;
static pthread_mutex_t request_db_mutex;
static int request_flag = 0;

/*****************************************************************************/
/*                                                                           */
#define NUMBER_COLUMN_CAT_ID     0
#define NUMBER_COLUMN_NAME       1
#define CATEGORY_NAME_QUERY_SIZE 50
static char category_name_query[CATEGORY_NAME_QUERY_SIZE] = "SELECT cat_id,name FROM cat WHERE locale='" LOCALE_NAME "'";

static int fill_category_name(void)
{
	int rc;
	sqlite3_stmt* stmt = NULL;
	rc = sqlite3_prepare_v2(database, category_name_query, -1, &stmt, NULL);
	if (rc != SQLITE_OK) {
		rc = DB_QUERY_CATEGORY_NAME_PREPARE;
		goto fill_category_name_exit;
	}
	for (;;) {
		rc = sqlite3_step(stmt);
		if (rc == SQLITE_ERROR) {
			rc = DB_QUERY_CATEGORY_NAME_STEP;
			break;
		}
		if (rc == SQLITE_DONE) {
			rc = SUCCESS;
			break;
		}
		if (rc != SQLITE_ROW) {
			rc = DB_QUERY_CATEGORY_NAME_NOT_ROW;
			break;
		}

		uint8_t type = sqlite3_column_int(stmt, NUMBER_COLUMN_CAT_ID);
		if (type == 0) {
			rc = DB_QUERY_CATEGORY_NAME_TYPE;
			break;
		}
		int byte = sqlite3_column_bytes(stmt, NUMBER_COLUMN_NAME);
		if (byte == 0) {
			rc = DB_QUERY_CATEGORY_NAME;
			break;
		}
		const unsigned char* name = sqlite3_column_text(stmt, NUMBER_COLUMN_NAME);
		rc = add_category_name(type, name, byte);
		if (rc != SUCCESS)
			break;
	}

fill_category_name_exit:
	sqlite3_finalize(stmt);
	return rc;
}

/*****************************************************************************/
#define QUERY_USER_VERSION_SIZE 37
static char query_user_version[] = "PRAGMA user_version = -1111111111111";
static char* query_user_version_p_number = query_user_version + 22;
#define QUERY_USER_VERSION_GET_SIZE    19
#define QUERY_USER_VERSION_NUMBER_SIZE 14

static int get_user_version(int32_t* user_version)
{
	int rc;
	*user_version = 0;

	sqlite3_stmt* stmt = NULL;
	rc = sqlite3_prepare_v2(database, query_user_version, QUERY_USER_VERSION_GET_SIZE, &stmt, NULL);
	if (rc != SQLITE_OK) {
		rc = DB_QUERY_USER_VERSION_GET_PREPARE;
		goto get_user_version_exit;
	}

	rc = sqlite3_step(stmt);
	if (rc == SQLITE_ERROR) {
		rc = DB_QUERY_USER_VERSION_GET_STEP;
		goto get_user_version_exit;
	}
	if (rc != SQLITE_ROW) {
		rc = DB_USER_VERSION_GET;
		goto get_user_version_exit;
	}
	rc = SUCCESS;
	const unsigned char* str_user_version = sqlite3_column_text(stmt, 0);
	*user_version = (int32_t)strtol((const char*)str_user_version, NULL, 10);

get_user_version_exit:
	sqlite3_finalize(stmt);
	return rc;
}

static int set_user_version(int32_t user_version)
{
	int rc;
	snprintf(query_user_version_p_number, QUERY_USER_VERSION_NUMBER_SIZE, "%d", user_version);

	sqlite3_stmt* stmt = NULL;
	rc = sqlite3_prepare_v2(database, query_user_version, -1, &stmt, NULL);
	if (rc != SQLITE_OK) {
		rc = DB_QUERY_USER_VERSION_SET_PREPARE;
		goto set_user_version_exit;
	}

	rc = sqlite3_step(stmt);
	if (rc == SQLITE_ERROR) {
		rc = DB_QUERY_USER_VERSION_SET_STEP;
		goto set_user_version_exit;
	}
	if (rc != SQLITE_DONE) {
		rc = DB_USER_VERSION_SET;
		goto set_user_version_exit;
	}
	rc = SUCCESS;

set_user_version_exit:
	sqlite3_finalize(stmt);
	return rc;
}

static int open_database(int32_t* user_version)
{
	int rc = sqlite3_open_v2(database_name, &database, SQLITE_OPEN_READWRITE, NULL);
	if (rc != SQLITE_OK)
		return DB_OPEN;

	/*                                */
	rc = sqlite3_exec(database, "PRAGMA JOURNAL_MODE = OFF", NULL, NULL, NULL);
	if (rc != SQLITE_OK)
		return DB_JOURNAL_MODE;

	sqlite3_exec(database, "PRAGMA TEMP_STORE = MEMORY", NULL, NULL, NULL);
	if (rc != SQLITE_OK)
		return DB_TEMP_STORE;

	sqlite3_exec(database, "PRAGMA SYNCHRONOUS = OFF", NULL, NULL, NULL);
	if (rc != SQLITE_OK)
		return DB_SYNCHRONOUS;

	sqlite3_exec(database, "PRAGMA LOCKING_MODE = EXCLUSIVE", NULL, NULL, NULL);
	if (rc != SQLITE_OK)
		return DB_LOCKING_MODE;

	rc = fill_category_name();
	if (rc != SUCCESS)
		return rc;
	rc = get_user_version(user_version);
	if (rc != SUCCESS)
		return rc;
	return SUCCESS;
}

static int close_database(int32_t* user_version)
{
	if (database != NULL) {
		sqlite3_close(database);
		database = NULL;
		*user_version = 0;
	}
	free_category_names();
	return SUCCESS;
}

/*****************************************************************************/
/*                                                                           */
static int32_t user_version = 0;
static char str_user_version[DATABASE_USER_VERSION_SIZE] = { 0 };

const char* get_user_version_database(void)
{
	snprintf(str_user_version, DATABASE_USER_VERSION_SIZE, "%08x", (unsigned int)user_version);
	return str_user_version;
}

/*****************************************************************************/
/*                                                                           */

struct requests
{
	/*                                                                           */
	char* update_header_buf;
	size_t update_header_buf_len;
	char* update_header_p_hash;

	/*                                                                           */
	char* update_full_buf;
	size_t update_full_buf_len;
	char* update_full_p_hash;

	/*                                                                           */
	char* download_header_buf;
	size_t download_header_buf_len;

	/*                                                                           */
	char* download_full_buf;
	size_t download_full_buf_len;
};
typedef struct requests requests_s;

static requests_s request = {
	.update_header_buf = NULL,
	.update_header_buf_len = 0,
	.update_header_p_hash = NULL,

	.update_full_buf = NULL,
	.update_full_buf_len = 0,
	.update_full_p_hash = NULL,

	.download_header_buf = NULL,
	.download_header_buf_len = 0,

	.download_full_buf = NULL,
	.download_full_buf_len = 0,
};

/*                                                                           */
static char REQUEST_END_STR[] = "\r\n\r\n";
#define REQUEST_END_STR_SIZE 4

/*                                                                           */
static char UPDATE_MAGIC_CONTENT[] = "PRAGMA user_version = 'xxxxxxxx';\n";
/*                                                                           */
/*                                  ,                                        */
static char UPDATE_FULL_MAGIC_CONTENT[] = "DELETE FROM 'result';";
/*                                                                           */
#define UPDATE_MAGIC_CONTENT_COMP_SIZE 23
/*                                                                              */
#define UPDATE_MAGIC_CONTENT_SIZE 34
/*                                                                           */
#define UPDATE_FULL_MAGIC_CONTENT_SIZE 21
/*                           0-54 == 55      (34 + 21)                       */
static char UPDATE_HEADER_RANGE_BYTE[] = "55";
/*                                                                           */
#define UPDATE_HEADER_RANGE_BYTE_LEN 2

static int init_request_update(char* host_str, char* port_str, char* path_str, char* user_str, char* password_str)
{
	size_t host_str_len = strlen(host_str);
	size_t port_str_len = strlen(port_str);
	size_t path_str_len = strlen(path_str);
	size_t user_str_len = strlen(user_str);
	size_t password_str_len = strlen(password_str);

	char credentials_str[(user_str_len + password_str_len + 1 + 1)];
	char* buf = credentials_str;
	memmove(buf, user_str, user_str_len);
	buf += user_str_len;
	*buf = ':';
	buf += 1;
	memmove(buf, password_str, password_str_len);
	uint8_t* authorization;
	size_t authorization_len;
	int rc = encode_base64((uint8_t*)credentials_str, (user_str_len + password_str_len + 1), &authorization, &authorization_len);
	if (rc != SUCCESS)
		return rc;

#define REQUEST_UPDATE_BASE_LEN (4 + 9 + 11 + 6 + 2 + 20 + 2 + 1 + 3 + 2 + 23 + 21 + 2 + 15 + 4)
	request.update_header_buf_len = REQUEST_UPDATE_BASE_LEN + path_str_len + host_str_len + authorization_len + UPDATE_HEADER_RANGE_BYTE_LEN;
	if (port_str_len != 0)
		request.update_header_buf_len += (port_str_len + 1);

	request.update_header_buf = malloc(request.update_header_buf_len + 1);
	if (request.update_header_buf == NULL) {
		free(authorization);
		return DB_REQUEST_UPDATE_HEAD_MALLOC_NULL;
	}
	request.update_full_buf_len = REQUEST_UPDATE_BASE_LEN + path_str_len + host_str_len + authorization_len;
	if (port_str_len != 0)
		request.update_full_buf_len += (port_str_len + 1);

	request.update_full_buf = malloc(request.update_full_buf_len + 1);
	if (request.update_full_buf == NULL) {
		free(authorization);
		free(request.update_header_buf);
		request.update_header_buf = NULL;
		return DB_REQUEST_UPDATE_FULL_MALLOC_NULL;
	}

	buf = request.update_header_buf;

	memmove(buf, "GET ", 4);
	buf += 4;
	memmove(buf, path_str, path_str_len);
	buf += path_str_len;
	memmove(buf, "/xxxxxxxx", 9);
	buf += 9;
	memmove(buf, " HTTP/1.1\r\n"
	             "Host: ",
	        (11 + 6));
	buf += (11 + 6);
	memmove(buf, host_str, host_str_len);
	buf += host_str_len;
	if (port_str_len != 0) {
		*buf = ':';
		buf += 1;
		memmove(buf, port_str, port_str_len);
		buf += port_str_len;
	}
	memmove(buf, "\r\n"
	             "User-Agent: url2cat-",
	        (2 + 20));
	buf += (2 + 20);
	char version_major[3] = { 0 };
	snprintf(version_major, 3, "%02d", PROJECT_VERSION_MAJOR);
	memmove(buf, version_major, 2);
	buf += 2;
	*buf = '.';
	buf += 1;
	char version_minor[4] = { 0 };
	snprintf(version_minor, 4, "%03d", PROJECT_VERSION_MINOR);
	memmove(buf, version_minor, 3);
	buf += 3;
	memmove(buf, "\r\n"
	             "Accept: application/*\r\n"
	             "Authorization: Basic ",
	        (2 + 23 + 21));
	buf += (2 + 23 + 21);
	memmove(buf, authorization, authorization_len);
	buf += authorization_len;
	memmove(buf, "\r\n"
	             "Range: bytes=0-",
	        (2 + 15));
	buf += (2 + 15);

	/*                                 */
	memmove(buf, REQUEST_END_STR, 4);
	memmove(request.update_full_buf, request.update_header_buf, request.update_full_buf_len);

	memmove(buf, UPDATE_HEADER_RANGE_BYTE, UPDATE_HEADER_RANGE_BYTE_LEN);
	buf += UPDATE_HEADER_RANGE_BYTE_LEN;
	memmove(buf, REQUEST_END_STR, 4);

	request.update_header_p_hash = request.update_header_buf + (4 + path_str_len + 1);
	request.update_full_p_hash = request.update_full_buf + (4 + path_str_len + 1);

	free(authorization);
	return SUCCESS;
}

static int deinit_request_update(void)
{
	if (request.update_header_buf != NULL) {
		free(request.update_header_buf);
		request.update_header_buf = NULL;
	}
	if (request.update_full_buf != NULL) {
		free(request.update_full_buf);
		request.update_full_buf = NULL;
	}
	return SUCCESS;
}

/*                                                                           */
static char DOWNLOAD_MAGIC_CONTENT[] = "SQLite format 3";
/*                                                                              */
#define DOWNLOAD_MAGIC_CONTENT_COMP_SIZE 15
/*                           0-14 == 15                                      */
static char DOWNLOAD_HEADER_RANGE_BYTE[] = "14";
/*                                                                           */
#define DOWNLOAD_HEADER_RANGE_BYTE_LEN 2

static int init_request_download(char* host_str, char* port_str, char* path_str, char* user_str, char* password_str)
{
	size_t host_str_len = strlen(host_str);
	size_t port_str_len = strlen(port_str);
	size_t path_str_len = strlen(path_str);
	size_t user_str_len = strlen(user_str);
	size_t password_str_len = strlen(password_str);

	char credentials_str[(user_str_len + password_str_len + 1 + 1)];
	char* buf = credentials_str;
	memmove(buf, user_str, user_str_len);
	buf += user_str_len;
	*buf = ':';
	buf += 1;
	memmove(buf, password_str, password_str_len);
	uint8_t* authorization;
	size_t authorization_len;
	int rc = encode_base64((uint8_t*)credentials_str, (user_str_len + password_str_len + 1), &authorization, &authorization_len);
	if (rc != SUCCESS)
		return rc;

#define REQUEST_DOWNLOAD_BASE_LEN (4 + 11 + 6 + 2 + 20 + 2 + 1 + 3 + 2 + 21 + 23 + 2 + 15 + 4)
	request.download_header_buf_len = REQUEST_DOWNLOAD_BASE_LEN + path_str_len + host_str_len + authorization_len + DOWNLOAD_HEADER_RANGE_BYTE_LEN;
	if (port_str_len != 0)
		request.download_header_buf_len += (port_str_len + 1);
	request.download_header_buf = malloc(request.download_header_buf_len + 1);
	if (request.download_header_buf == NULL) {
		free(authorization);
		return DB_REQUEST_DOWNLOAD_HEAD_MALLOC_NULL;
	}
	request.download_full_buf_len = REQUEST_DOWNLOAD_BASE_LEN + path_str_len + host_str_len + authorization_len;
	if (port_str_len != 0)
		request.download_full_buf_len += (port_str_len + 1);
	request.download_full_buf = malloc(request.download_full_buf_len + 1);
	if (request.download_full_buf == NULL) {
		free(authorization);
		free(request.download_header_buf);
		request.download_header_buf = NULL;
		return DB_REQUEST_DOWNLOAD_FULL_MALLOC_NULL;
	}

	buf = request.download_header_buf;

	memmove(buf, "GET ", 4);
	buf += 4;
	memmove(buf, path_str, path_str_len);
	buf += path_str_len;
	memmove(buf, " HTTP/1.1\r\n"
	             "Host: ",
	        (11 + 6));
	buf += (11 + 6);
	memmove(buf, host_str, host_str_len);
	buf += host_str_len;
	if (port_str_len != 0) {
		*buf = ':';
		buf += 1;
		memmove(buf, port_str, port_str_len);
		buf += port_str_len;
	}
	memmove(buf, "\r\n"
	             "User-Agent: url2cat-",
	        (2 + 20));
	buf += (2 + 20);
	char version_major[3] = { 0 };
	snprintf(version_major, 3, "%02d", PROJECT_VERSION_MAJOR);
	memmove(buf, version_major, 2);
	buf += 2;
	*buf = '.';
	buf += 1;
	char version_minor[4] = { 0 };
	snprintf(version_minor, 4, "%03d", PROJECT_VERSION_MINOR);
	memmove(buf, version_minor, 3);
	buf += 3;
	memmove(buf, "\r\n"
	             "Accept: application/*\r\n"
	             "Authorization: Basic ",
	        (2 + 23 + 21));
	buf += (2 + 23 + 21);
	memmove(buf, authorization, authorization_len);
	buf += authorization_len;
	memmove(buf, "\r\n"
	             "Range: bytes=0-",
	        (2 + 15));
	buf += (2 + 15);

	/*                                    */
	memmove(buf, REQUEST_END_STR, 4);
	memmove(request.download_full_buf, request.download_header_buf, request.download_full_buf_len);

	memmove(buf, DOWNLOAD_HEADER_RANGE_BYTE, DOWNLOAD_HEADER_RANGE_BYTE_LEN);
	buf += DOWNLOAD_HEADER_RANGE_BYTE_LEN;
	memmove(buf, REQUEST_END_STR, 4);

	free(authorization);
	return SUCCESS;
}

static int deinit_request_download(void)
{
	if (request.download_header_buf != NULL) {
		free(request.download_header_buf);
		request.download_header_buf = NULL;
	}
	if (request.download_full_buf != NULL) {
		free(request.download_full_buf);
		request.download_full_buf = NULL;
	}
	return SUCCESS;
}

/******************************************************************************/
/*                                                                            */
static char ANSWER_PARTIAL_CONTENT[] = "HTTP/1.1 206 Partial Content";
#define ANSWER_PARTIAL_CONTENT_SIZE 28
static char CONTENT_RANGE_STR[] = "Content-Range: bytes 0-";
#define CONTENT_RANGE_STR_SIZE 23
#if (UPDATE_HEADER_RANGE_BYTE_LEN == DOWNLOAD_HEADER_RANGE_BYTE_LEN)
	#define HEADER_RANGE_BYTE_LEN UPDATE_HEADER_RANGE_BYTE_LEN
#else
	#error "Change parse_answer_header"
#endif

static int parse_answer_header(char* buffer, size_t size, size_t* file_size, size_t* header_size, char** file_begin)
{
	int rc;
	/*            HTTP           */
	rc = strncmp(buffer, ANSWER_PARTIAL_CONTENT, ANSWER_PARTIAL_CONTENT_SIZE);
	if (rc != 0)
		return DB_PARSE_ANSWER_NOT_OK;

	char* len_str = strstr(buffer, CONTENT_RANGE_STR);
	if (len_str == NULL)
		return DB_PARSE_ANSWER_NOT_CONTENT_RANGE;

	len_str += CONTENT_RANGE_STR_SIZE + HEADER_RANGE_BYTE_LEN + 1;

	/*                             */
	char* header_str = strstr(buffer, REQUEST_END_STR);
	if (header_str == NULL)
		return DB_PARSE_ANSWER_NOT_END_HEAD;

	header_str = header_str + REQUEST_END_STR_SIZE;
	*header_size = header_str - buffer;
	*file_begin = header_str;

	for (char* len_str_end = len_str;; ++len_str_end) {
		if ((*len_str_end == '\r') || (*len_str_end == '\n') || (*len_str_end == 0)) {
			*len_str_end = 0;
			break;
		}
	}

	size_t len = strtol(len_str, NULL, 10);
	if (len == 0)
		return DB_PARSE_ANSWER_NOT_LENGTH;
	*file_size = len;
	/*                                                   */
	size_t len_str_len = strlen(len_str);
	*header_size += ((len_str_len - HEADER_RANGE_BYTE_LEN) << 1);
	return SUCCESS;
}

static int parse_answer_update(char* begin_file, int32_t* user_version_new, int* update)
{
	int rc = strncmp(begin_file, UPDATE_MAGIC_CONTENT, UPDATE_MAGIC_CONTENT_COMP_SIZE);
	if (rc != 0)
		return DB_UPDATE_NOT_MAGIC_STR_HEAD;
	char* p_version = begin_file + UPDATE_MAGIC_CONTENT_COMP_SIZE;
	int32_t version = (int32_t)strtol(p_version, NULL, 16);
	if (version == 0)
		return DB_UPDATE_NOT_USER_VERSION_HEAD;
	*user_version_new = version;

	/*                                                    */
	char* p_update = begin_file + UPDATE_MAGIC_CONTENT_SIZE;
	rc = strncmp(p_update, UPDATE_FULL_MAGIC_CONTENT, UPDATE_FULL_MAGIC_CONTENT_SIZE);
	if (rc == 0)
		*update = NO;
	else
		*update = YES;
	return SUCCESS;
}

static int parse_answer_download(char* begin_file)
{
	int rc = strncmp(begin_file, DOWNLOAD_MAGIC_CONTENT, DOWNLOAD_MAGIC_CONTENT_COMP_SIZE);
	if (rc != 0)
		return DB_DOWNLOAD_NOT_MAGIC_STR_HEAD;
	return SUCCESS;
}

/******************************************************************************/
/*                                                                            */
static char* file_buffer = NULL;
static size_t file_buffer_size = 0;
static int update_flag = NO;
static int fd_update = -1;

static char name_temp[] = "db_xxxxxxxx.diff.sql";
#define NAME_TEMP_STR_SIZE 21

#define DELAY_CHECK_REQUEST 200

static int remove_download_file(int reload_database)
{
	if (file_buffer != NULL) {
		munmap(file_buffer, file_buffer_size);
		file_buffer = NULL;
		file_buffer_size = 0;
	}

	if (fd_update != -1) {
		close(fd_update);
		fd_update = -1;
	}

	if (reload_database == NO) {
		unlink(name_temp);
	}
	else {
		/*                        */

		pthread_mutex_lock(&update_db_mutex);
		/* check reqests*/
		for (;;) {
			pthread_mutex_lock(&request_db_mutex);
			if (request_flag == 0) {
				pthread_mutex_unlock(&request_db_mutex);
				break;
			}
			pthread_mutex_unlock(&request_db_mutex);
			usleep(DELAY_CHECK_REQUEST);
		}

		int rc = close_database(&user_version);
		if (rc != SUCCESS) {
			pthread_mutex_unlock(&update_db_mutex);
			return rc;
		}
		rc = rename(name_temp, database_name);
		if (rc == -1) {
			pthread_mutex_unlock(&update_db_mutex);
			return DB_FILE_RENAME;
		}
		rc = open_database(&user_version);
		pthread_mutex_unlock(&update_db_mutex);
		if (rc != SUCCESS) {
			return rc;
		}
	}

	return SUCCESS;
}

static int download_file(int32_t user_version_new, size_t file_size, net_connect_s* net_connect)
{
	int rc;
	snprintf(name_temp, NAME_TEMP_STR_SIZE, "db_%08x.diff.sql", (uint32_t)user_version_new);
	int fd_update = open(name_temp, O_CREAT | O_TRUNC | O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd_update == -1) {
		rc = DB_DOWNLOAD_FILE_NOT_CREAT;
		goto download_file_exit;
	}
	struct statfs fs_stat;
	rc = statfs(name_temp, &fs_stat);
	if (rc != 0) {
		rc = DB_DOWNLOAD_FILE_NOT_FS_STAT;
		goto download_file_exit;
	}
	size_t file_size_check = fs_stat.f_bavail * fs_stat.f_bsize;
	if (file_size_check < file_size) {
		rc = DB_DOWNLOAD_FILE_NOT_SPACE_DEVICE;
		goto download_file_exit;
	}

	file_size_check = lseek(fd_update, (file_size - 1), SEEK_SET);
	if ((file_size - 1) != file_size_check) {
		rc = DB_DOWNLOAD_FILE_NOT_FULL;
		goto download_file_exit;
	}

	rc = write(fd_update, "0", 1);
	if (rc != 1) {
		rc = DB_DOWNLOAD_FILE_NOT_WRITE;
		goto download_file_exit;
	}

	file_buffer = mmap(0, file_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd_update, 0);
	if (file_buffer == MAP_FAILED) {
		rc = DB_DOWNLOAD_FILE_NOT_MMAP;
		file_buffer = NULL;
		goto download_file_exit;
	}
	file_buffer_size = file_size;

	char* current = file_buffer;
	file_size_check = 0;
#define SIZE_FILE_PART 0x10000
	size_t size = 0;
#if SHOW_DOWNLOAD
	int check = 1000;
	printf("\ndownload : 00.0%%");
	fflush(stdout);
#endif
	for (; file_size_check < file_size;) {
		size = recv_net(net_connect, current, SIZE_FILE_PART);
		if (size <= 0) {
#if SHOW_DOWNLOAD
			printf("\n");
#endif
			rc = DB_DOWNLOAD_FILE_READ;
			goto download_file_exit;
		}
		file_size_check += size;
		current += size;
#if SHOW_DOWNLOAD
		--check;
		if (check == 0) {
			check = 1000;
			size = file_size_check * 1000;
			size = size / file_size;
			int d = size / 10;
			int e = size - (d * 10);
			printf("\rdownload : %02d.%01d%%", d, e);
			fflush(stdout);
		}
#endif
	}
	rc = SUCCESS;
#if SHOW_DOWNLOAD
	printf("\rdownload :  100%%\n");
#endif
download_file_exit:
	if (rc != SUCCESS) {
		remove_download_file(NO);
	}
	return rc;
}

/******************************************************************************/
/*                                                                            */

static int apply_patch_query(char* req_str, size_t req_str_len)
{
	sqlite3_stmt* stmt = NULL;

	int rc = sqlite3_prepare_v2(database, req_str, req_str_len, &stmt, NULL);
	if (rc != SQLITE_OK) {
		rc = DB_QUERY_APPLY_PATCH_PREPARE;
		goto apply_patch_query_exit;
	}
	rc = sqlite3_step(stmt);
	if (rc == SQLITE_ERROR) {
		rc = DB_QUERY_APPLY_PATCH_STEP;
		goto apply_patch_query_exit;
	}
	if (rc != SQLITE_DONE) {
		rc = DB_QUERY_APPLY_PATCH_NOT_DONE;
		goto apply_patch_query_exit;
	}
	rc = SUCCESS;
apply_patch_query_exit:
	sqlite3_finalize(stmt);
	return rc;
}

static int apply_patch(int32_t user_version_new)
{
	char* buffer = file_buffer;
	size_t buffer_size = file_buffer_size;

	int rc = strncmp(buffer, UPDATE_MAGIC_CONTENT, UPDATE_MAGIC_CONTENT_COMP_SIZE);
	if (rc != 0)
		return DB_APPLY_PATCH_NOT_MAGIC_STR;

	buffer += UPDATE_MAGIC_CONTENT_COMP_SIZE;
	long int version = strtol(buffer, NULL, 16);
	if (version == 0)
		return DB_APPLY_PATCH_NOT_USER_VERSION;

	if (user_version_new != ((int32_t)version))
		return DB_APPLY_PATCH_MISTMACH_USER_VERSION;

	char* begin = buffer;
	char* end = NULL;
	size_t size = UPDATE_MAGIC_CONTENT_COMP_SIZE;
	end = memchr(begin, ';', (buffer_size - size));
	if (end == NULL)
		return DB_APPLY_PATCH_NOT_END_ROW;

	for (;;) {
		if ((*end == ';')
		    || (*end == '\n')
		    || (*end == '\r')
		    || (*end == '\t')
		    || (*end == ' '))
			++end;
		else
			break;
	}
	/*                                   user_version */
	size += end - begin;
	begin = end;
#if SHOW_DOWNLOAD
	printf("\nupdate : 00.0%%");
	fflush(stdout);
#endif
	for (; size < buffer_size;) {
		end = memchr(begin, ';', (buffer_size - size));
		if (end == NULL)
			return DB_APPLY_PATCH_NOT_END_ROW;

		for (;;) {
			if ((*end == ';')
			    || (*end == '\n')
			    || (*end == '\r')
			    || (*end == '\t')
			    || (*end == ' '))
				++end;
			else
				break;
		}
		size_t len = end - begin;
		rc = apply_patch_query(begin, len);
		if (rc != SUCCESS)
			return rc;
		size += len;
		begin = end;
#if SHOW_DOWNLOAD
		len = size * 1000;
		len = len / buffer_size;
		int d = len / 10;
		int e = len - (d * 10);
		printf("\rupdate : %02d.%01d%%", d, e);
		fflush(stdout);
#endif
	}
#if SHOW_DOWNLOAD
	printf("\n");
#endif

	rc = set_user_version(user_version_new);
	if (rc != SUCCESS)
		return rc;

	return SUCCESS;
}

/******************************************************************************/
/*                                                                            */

static int download_database(net_connect_s* net_connect, char* answer_buffer, int32_t user_version_new)
{
	/*                           */
#if SHOW_DOWNLOAD
	*(request.download_header_buf + request.download_header_buf_len) = 0;
	printf("download request:\n%s\n", request.download_header_buf);
#endif
	int rc = send_net(net_connect, request.download_header_buf, request.download_header_buf_len);
	if (rc != request.download_header_buf_len)
		return DB_DOWNLOAD_HEADER_WRITE;

	/*                                           */
	rc = recv_net(net_connect, answer_buffer, (ANSWER_BUFFER_SIZE_DEFAULT - 1));
	if (rc <= 0)
		return DB_DOWNLOAD_HEADER_READ;
	*(answer_buffer + rc) = 0;
#if SHOW_DOWNLOAD
	printf("download answer:\n%s\n", answer_buffer);
#endif
	size_t file_size;
	size_t header_size;
	char* file_begin;
	/*                       ,                                 */
	rc = parse_answer_header(answer_buffer, rc, &file_size, &header_size, &file_begin);
	if (rc != SUCCESS)
		return rc;

	/*                       */
	rc = parse_answer_download(file_begin);
	if (rc != SUCCESS)
		return rc;

	/*                                     */
	rc = send_net(net_connect, request.download_full_buf, request.download_full_buf_len);
	if (rc != request.download_full_buf_len)
		return DB_DOWNLOAD_FULL_WRITE;
	/*                            */
	rc = recv_net(net_connect, answer_buffer, header_size);
	if (rc <= 0)
		return DB_DOWNLOAD_FULL_READ;

	/*               */
	rc = download_file(user_version_new, file_size, net_connect);
	if (rc != SUCCESS)
		return rc;

	return SUCCESS;
}

static net_option_s net_option = { 0 };
static net_connect_s net_connect = { 0 };

int update_database(void)
{
	int rc;
	int reload_database = NO;

	if (update_flag == YES)
		return DB_IS_BEING_UPDATED;
	update_flag = YES;

	rc = connect_net(&net_option, &net_connect);
	if (rc != SUCCESS)
		return rc;

	/*                                 */
	const char* user_version_hash = get_user_version_database();
	memmove(request.update_header_p_hash, user_version_hash, DATABASE_USER_VERSION_SIZE - 1);
	memmove(request.update_full_p_hash, user_version_hash, DATABASE_USER_VERSION_SIZE - 1);

#if SHOW_DOWNLOAD
	*(request.update_header_buf + request.update_header_buf_len) = 0;
	printf("update request:\n%s\n", request.update_header_buf);
#endif
	/*                                   */
	rc = send_net(&net_connect, request.update_header_buf, request.update_header_buf_len);
	if (rc != request.update_header_buf_len) {
		rc = DB_UPDATE_HEADER_WRITE;
		goto update_database_exit;
	}

	char answer_buffer[ANSWER_BUFFER_SIZE_DEFAULT];

	/*                                  */
	rc = recv_net(&net_connect, answer_buffer, (ANSWER_BUFFER_SIZE_DEFAULT - 1));
	if (rc <= 0) {
		rc = DB_UPDATE_HEADER_READ;
		goto update_database_exit;
	}
	*(answer_buffer + rc) = 0;
#if SHOW_DOWNLOAD
	printf("update answer:\n%s\n", answer_buffer);
#endif
	size_t file_size;
	size_t header_size;
	char* file_begin;
	rc = parse_answer_header(answer_buffer, rc, &file_size, &header_size, &file_begin);
	if (rc != SUCCESS)
		goto update_database_exit;

	int32_t user_version_new = 0;
	int update;
	rc = parse_answer_update(file_begin, &user_version_new, &update);
	if (rc != SUCCESS)
		goto update_database_exit;

	/*                                         */
	if (user_version_new == user_version) {
		rc = SUCCESS;
		goto update_database_exit;
	}

	if (update == YES) {
		/*                           */
		rc = send_net(&net_connect, request.update_full_buf, request.update_full_buf_len);
		if (rc != request.update_full_buf_len) {
			rc = DB_UPDATE_FULL_WRITE;
			goto update_database_exit;
		}
		/*                            */
		rc = recv_net(&net_connect, answer_buffer, header_size);
		if (rc <= 0) {
			rc = DB_UPDATE_FULL_READ;
			goto update_database_exit;
		}

		rc = download_file(user_version_new, file_size, &net_connect);
		if (rc != SUCCESS)
			goto update_database_exit;

		/* apply patch */
		pthread_mutex_lock(&update_db_mutex);
		/* check reqests*/
		for (;;) {
			pthread_mutex_lock(&request_db_mutex);
			if (request_flag == 0) {
				pthread_mutex_unlock(&request_db_mutex);
				break;
			}
			pthread_mutex_unlock(&request_db_mutex);
			usleep(DELAY_CHECK_REQUEST);
		}

		rc = apply_patch(user_version_new);
		pthread_mutex_unlock(&update_db_mutex);
		if (rc != SUCCESS)
			goto update_database_exit;
		user_version = user_version_new;
		rc = SUCCESS;
	}
	else { /*update == NO */
		rc = download_database(&net_connect, answer_buffer, user_version_new);
		if (rc != SUCCESS)
			goto update_database_exit;
		reload_database = YES;
	}

update_database_exit:
	remove_download_file(reload_database);
	disconnect_net(&net_connect);
	update_flag = NO;
	return rc;
}

/*****************************************************************************/
/*                                                                           */

static char* category_query_buf = NULL;
static char* category_query_p_domain = NULL;
static char* category_query_p_path = NULL;

/* Query                                                                     */
/* SELECT cat_id FROM result WHERE domain_hash=x'hash' AND path_hash=x'hash' */
#define CATEGORY_QUERY_BASE_SIZE (46 + 18 + 1)
size_t category_query_size = CATEGORY_QUERY_BASE_SIZE + (DOMAIN_HASH_SIZE << 1) + (PATH_HASH_SIZE << 1);

static int init_category_query(void)
{
	category_query_buf = malloc((category_query_size + 1));
	if (category_query_buf == NULL)
		return DB_QUERY_CATEGORY_MALLOC_NULL;

	char* buf = category_query_buf;
	memmove(buf, "SELECT cat_id FROM result WHERE domain_hash=x'", 46);
	buf += 46;
	category_query_p_domain = buf;
	memset(category_query_p_domain, 'x', (DOMAIN_HASH_SIZE << 1));
	buf += (DOMAIN_HASH_SIZE << 1);
	memmove(buf, "' AND path_hash=x'", 18);
	buf += 18;
	category_query_p_path = buf;
	memset(category_query_p_path, 'x', (PATH_HASH_SIZE << 1));
	buf += (PATH_HASH_SIZE << 1);
	*buf = '\'';

	return SUCCESS;
}

static int deinit_category_query(void)
{
	if (category_query_buf != NULL) {
		free(category_query_buf);
		category_query_buf = NULL;
		category_query_p_domain = NULL;
		category_query_p_path = NULL;
	}
	return SUCCESS;
}

/*****************************************************************************/
/*                                                                           */

static int execute_category_query(uint8_t hash_domain[DOMAIN_HASH_SIZE], uint8_t hash_path[PATH_HASH_SIZE], url2cat_category_s* p_category, int* p_number)
{
	char* str_hash = category_query_p_domain;
	for (int i = 0; i < DOMAIN_HASH_SIZE; ++i) {
		*str_hash = up_hex_char(hash_domain[i]);
		++str_hash;
		*str_hash = low_hex_char(hash_domain[i]);
		++str_hash;
	}

	int len = category_query_size;
	str_hash = category_query_p_path;
	if (hash_path[0] == 0) {
		*str_hash = '\'';
		len = category_query_size - (PATH_HASH_SIZE << 1);
	}
	else {
		for (int i = 0; i < PATH_HASH_SIZE; ++i) {
			*str_hash = up_hex_char(hash_path[i]);
			++str_hash;
			*str_hash = low_hex_char(hash_path[i]);
			++str_hash;
		}
	}
	static sqlite3_stmt* stmt = NULL;
	int rc = sqlite3_prepare_v2(database, category_query_buf, len, &stmt, NULL);
	if (rc != SQLITE_OK) {
		rc = DB_QUERY_CATEGORY_PREPARE;
		goto execute_category_query_exit;
	}
	rc = sqlite3_step(stmt);
	if (rc == SQLITE_ERROR) {
		rc = DB_QUERY_CATEGORY_STEP;
		goto execute_category_query_exit;
	}
	if (rc == SQLITE_DONE) {
		rc = DB_URL_NOT_FOUND;
		goto execute_category_query_exit;
	}
	if (rc != SQLITE_ROW) {
		rc = DB_QUERY_CATEGORU_NOT_ROW;
		goto execute_category_query_exit;
	}

	int byte = sqlite3_column_bytes(stmt, 0);
	unsigned short* cat = (unsigned short*)sqlite3_column_blob(stmt, NUMBER_COLUMN_CAT_ID);
	byte >>= 1;
	int number;
	for (number = 0; ((number < byte) && (number < *p_number)); ++number) {
		unsigned short type = *cat;
		type >>= 8;
		p_category->type = type;
		const char* name = get_category_name(type);
		if (name != NULL)
			strncpy(p_category->type_name, name, CATEGORY_TYPE_NAME_STR_SIZE);
		else
			p_category->type_name[0] = 0;
		++p_category;
		++cat;
	}
	if (*p_number > number)
		*p_number = number;

	rc = SUCCESS;
execute_category_query_exit:
	sqlite3_finalize(stmt);
	return rc;
}

int get_category_database(char* url_str, size_t url_str_size, url2cat_category_s* p_category, int* p_number)
{
	char* domain;
	char* path;
	int rc;

	if (database == NULL)
		return DB_URL_NOT_FOUND;

	domain = url_str;
	path = memchr(url_str, '/', url_str_size);
	if (path == NULL)
		return DB_URL_NO_PATH;

	rc = pthread_mutex_trylock(&update_db_mutex);
	if (rc == EBUSY)
		return DB_IS_BEING_UPDATED;
	pthread_mutex_unlock(&update_db_mutex);

	pthread_mutex_lock(&request_db_mutex);
	++request_flag;
	pthread_mutex_unlock(&request_db_mutex);

	/*                           */
	/* /                                    */
#if defined(HASH_MD5)
	size_t len_domain = path - url_str;
	size_t len_path = url_str_size - len_domain;
#elif defined(HASH_SHA256) || defined(HASH_SHA512)
	size_t len_domain = path - url_str + 1;
	size_t len_path = url_str_size - len_domain + 1;
#endif

	uint8_t* hash = COMPUTES_HASH((const uint8_t*)domain, len_domain, NULL);
	uint8_t hash_domain[DOMAIN_HASH_SIZE];
	memmove(hash_domain, hash, DOMAIN_HASH_SIZE);
	uint8_t hash_path[PATH_HASH_SIZE];
	if (len_path > sizeof(char)) {
		hash = COMPUTES_HASH((const uint8_t*)path, len_path, NULL);
		memmove(hash_path, hash, PATH_HASH_SIZE);
	}
	else {
		hash_path[0] = 0;
	}
	rc = execute_category_query(hash_domain, hash_path, p_category, p_number);
	if ((rc != SUCCESS) && (rc != DB_URL_NOT_FOUND)) {
		close_database(&user_version);
	}

	pthread_mutex_lock(&request_db_mutex);
	--request_flag;
	pthread_mutex_unlock(&request_db_mutex);

	return rc;
}

/******************************************************************************/
/*                                                                            */
int init_database(char* name, char* scheme_str, char* host_str, char* port_str, char* update_path_str, char* download_path_str,
                  char* user_str, char* password_str)
{
	int rc;

	strncpy(database_name, name, DATABASE_NAME_STR_SIZE);

	rc = init_category_query();
	if (rc != SUCCESS)
		return rc;

	rc = init_request_update(host_str, port_str, update_path_str, user_str, password_str);
	if (rc != SUCCESS) {
		deinit_category_query();
		return rc;
	}

	rc = init_request_download(host_str, port_str, download_path_str, user_str, password_str);
	if (rc != SUCCESS) {
		deinit_category_query();
		deinit_request_update();
		return rc;
	}

	rc = init_net(&net_option, scheme_str, host_str, port_str);
	if (rc != SUCCESS) {
		deinit_category_query();
		deinit_request_update();
		deinit_request_download();
		return rc;
	}

	rc = open_database(&user_version);
	if (rc != SUCCESS) {
		rc = connect_net(&net_option, &net_connect);
		if (rc != SUCCESS)
			goto download_database_exit;

		char answer_buffer[ANSWER_BUFFER_SIZE_DEFAULT];
		rc = download_database(&net_connect, answer_buffer, 0);
		if (rc == SUCCESS) {
			remove_download_file(YES);
			disconnect_net(&net_connect);
		}
		else {
		download_database_exit:
			deinit_category_query();
			deinit_request_update();
			deinit_request_download();
			remove_download_file(NO);
			disconnect_net(&net_connect);
			deinit_net(&net_option);
			close_database(&user_version);
			return rc;
		}
	}

	pthread_mutex_init(&update_db_mutex, NULL);
	pthread_mutex_init(&request_db_mutex, NULL);

	return SUCCESS;
}

int deinit_database(void)
{
	deinit_category_query();
	deinit_request_update();
	deinit_request_download();
	deinit_net(&net_option);
	close_database(&user_version);
	pthread_mutex_destroy(&update_db_mutex);
	pthread_mutex_destroy(&request_db_mutex);

	return SUCCESS;
}

/*****************************************************************************/
