/*****************************************************************************/
/*                                                                           */
/* liburl2cat : Domain Categorization Library                                */
/*                                                                           */
/*  Copyright (C) 2020 Kuzmin Yaroslav <kuzmin.yaroslav@gmail.com>           */
/*                                                                           */
/* liburl2cat is free software: you can redistribute it and/or modify it     */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation, either version 3 of the License, or             */
/* (at your option) any later version.                                       */
/*                                                                           */
/* liburl2cat is distributed in the hope that it will be useful, but         */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      */
/* See the GNU General Public License for more details.                      */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program.  If not, see <http://www.gnu.org/licenses/>.           */
/*                                                                           */
/*****************************************************************************/

#include "common.h"

#define SHOW_NET 0

/*****************************************************************************/
/* HTTPS support                                                             */
static int connect_net_https(net_option_s* net_option, net_connect_s* net_connect)
{
	int rc;
	SSL_CTX* ssl_ctx = NULL;
	SSL_CONF_CTX* ssl_conf_ctx = NULL;
	BIO* bio_host = NULL;

	const SSL_METHOD* ssl_method = TLS_client_method();
	ssl_ctx = SSL_CTX_new(ssl_method);
	ssl_conf_ctx = SSL_CONF_CTX_new();
	SSL_CONF_CTX_set_flags(ssl_conf_ctx, SSL_CONF_FLAG_CLIENT);
	SSL_CONF_CTX_set_ssl_ctx(ssl_conf_ctx, ssl_ctx);
	rc = SSL_CONF_CTX_finish(ssl_conf_ctx);
	if (rc == 0) {
		ERR_print_errors_fp(stderr);
		rc = NET_SSL_CONF_CTX_finish;
		goto net_https_connect_failure;
	}
	bio_host = BIO_new_ssl_connect(ssl_ctx);
	SSL* ssl = NULL;
	BIO_get_ssl(bio_host, &ssl);
	if (ssl == NULL) {
		ERR_print_errors_fp(stderr);
		rc = NET_SSL_BIO_get_ssl;
		goto net_https_connect_failure;
	}
	/*           */
	SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);

	BIO_set_conn_hostname(bio_host, net_option->host);
	rc = BIO_do_connect(bio_host);
	if (rc <= 0) {
		ERR_print_errors_fp(stderr);
		rc = NET_SSL_BIO_do_connect;
		goto net_https_connect_failure;
	}

	/*                    */
	BIO_set_nbio(bio_host, 1);
	int fd = BIO_get_fd(bio_host, NULL);
	int flag = fcntl(fd, F_GETFL);
	if (!(flag & O_NONBLOCK)) {
		fcntl(fd, F_SETFL, O_NONBLOCK);
	}

	rc = BIO_do_handshake(bio_host);
	if (rc <= 0) {
		ERR_print_errors_fp(stderr);
		rc = NET_SSL_BIO_do_handshake;
		goto net_https_connect_failure;
	}
	net_connect->ssl_ctx = ssl_ctx;
	net_connect->ssl_conf_ctx = ssl_conf_ctx;
	net_connect->bio_host = bio_host;
	return SUCCESS;
net_https_connect_failure:
	SSL_CONF_CTX_free(ssl_conf_ctx);
	SSL_CTX_free(ssl_ctx);
	BIO_free_all(bio_host);
	ssl_ctx = NULL;
	ssl_conf_ctx = NULL;
	bio_host = NULL;
	net_connect->ssl_ctx = NULL;
	net_connect->ssl_conf_ctx = NULL;
	net_connect->bio_host = NULL;
	return rc;
}

static int disconnect_net_https(net_connect_s* net_connect)
{
	if (net_connect->ssl_conf_ctx != NULL) {
		SSL_CONF_CTX_free(net_connect->ssl_conf_ctx);
		net_connect->ssl_conf_ctx = NULL;
	}
	if (net_connect->ssl_ctx != NULL) {
		SSL_CTX_free(net_connect->ssl_ctx);
		net_connect->ssl_ctx = NULL;
	}
	if (net_connect->bio_host != NULL) {
		BIO_free_all(net_connect->bio_host);
		net_connect->bio_host = NULL;
	}
	return SUCCESS;
}

static int send_net_https(net_connect_s* net_connect, char* buffer, size_t size)
{
	/*                                  */
	int rc = BIO_write(net_connect->bio_host, buffer, size);
	if (rc != size) {
		ERR_print_errors_fp(stderr);
	}
	return rc;
}
#define DELAY_MAX   30      /*30 second     */
#define DELAY_PART  1000000 /*0.001 second  */
#define MAX_NANOSEC 999999999
struct timespec time_delay = { .tv_sec = 0, .tv_nsec = DELAY_PART };
struct timespec time_max = { .tv_sec = 0, .tv_nsec = 0 };

static int recv_net_https(net_connect_s* net_connect, char* buffer, size_t size)
{
	int len;
	time_max.tv_sec = 0;
	time_max.tv_nsec = 0;
	/*                                  */
	for (;;) {
		len = BIO_read(net_connect->bio_host, buffer, size);
		if (len <= 0) {
			if (time_max.tv_sec > DELAY_MAX)
				return -1;
			nanosleep(&time_delay, NULL);
			time_max.tv_nsec += DELAY_PART;
			if (time_max.tv_nsec > MAX_NANOSEC) {
				time_max.tv_nsec = 0;
				++time_max.tv_sec;
#if SHOW_NET
				printf("timeout : %ld\n", time_max.tv_sec);
#endif
			}
			continue;
		}
		time_max.tv_sec = 0;
		break;
	}
	return len;
}

/*****************************************************************************/

struct support_scheme
{
	char* name;
	scheme_e type;
	uint32_t port;
};
typedef struct support_scheme support_scheme_s;

static support_scheme_s support_scheme[] = {
	{ .name = "https", .type = HTTPS, .port = 443 },
	{ .name = NULL, .type = 0, .port = 0 }
};

static support_scheme_s* check_scheme(char* scheme_str)
{
	int rc;
	support_scheme_s* current_scheme = support_scheme;
	for (;;) {
		if (current_scheme->name == NULL) {
			current_scheme = NULL;
			break;
		}
		size_t len = strlen(current_scheme->name);
		rc = strncmp(scheme_str, current_scheme->name, len);
		if (rc == 0) {
			size_t len_check = strlen(scheme_str);
			if (len == len_check)
				break;
		}
		++current_scheme;
	}
	return current_scheme;
}

static char port_check_str[NET_PORT_STR_SIZE] = { 0 };

int init_net(net_option_s* net_option, char* scheme_str, char* host_str, char* port_str)
{
	support_scheme_s* support_scheme = check_scheme(scheme_str);
	if (support_scheme == NULL)
		return NET_SCHEME_NOT_SUPPORT;
	long int port = 0;
	if (port_str != NULL)
		port = strtol(port_str, NULL, 10);
	if (port == 0)
		port = support_scheme->port;
	sprintf(port_check_str, "%ld", port);

	int len_host_str = strlen(host_str);
	int len_port_str = strlen(port_check_str);

	char* host = calloc(sizeof(char), (len_host_str + len_port_str + 2));
	if (host == NULL)
		return NET_HOST_PORT_MALLOC_NULL;
	strcat(host, host_str);
	strcat(host, ":");
	strcat(host, port_check_str);

	net_option->host = host;
	net_option->scheme = support_scheme->type;
	return SUCCESS;
}

int deinit_net(net_option_s* net_option)
{
	if (net_option->host != NULL) {
		free(net_option->host);
		net_option->host = NULL;
	}
	return SUCCESS;
}

int connect_net(net_option_s* net_option, net_connect_s* net_connect)
{
	switch (net_option->scheme) {
		case HTTPS:
			return connect_net_https(net_option, net_connect);
		default:
			return NET_CONNECT_FAILURE;
	}
}

int disconnect_net(net_connect_s* net_connect)
{
	if (net_connect->bio_host)
		disconnect_net_https(net_connect);

	return SUCCESS;
}

int send_net(net_connect_s* net_connect, char* buffer, size_t size)
{
	if (net_connect->bio_host)
		return send_net_https(net_connect, buffer, size);
	return NET_WRITE_FAILURE;
}

int recv_net(net_connect_s* net_connect, char* buffer, size_t size)
{
	if (net_connect->bio_host)
		return recv_net_https(net_connect, buffer, size);
	return NET_READ_FAILURE;
}

int free_net(void)
{
	EVP_cleanup();
	SSL_COMP_free_compression_methods();
	COMP_zlib_cleanup();
	CONF_modules_free();
	CONF_modules_unload(1);
	ERR_free_strings();
	CRYPTO_cleanup_all_ex_data();
	return SUCCESS;
}

/*****************************************************************************/
