/*****************************************************************************/
/*                                                                           */
/* liburl2cat : Domain Categorization Library                                */
/*                                                                           */
/*  Copyright (C) 2020 Kuzmin Yaroslav <kuzmin.yaroslav@gmail.com>           */
/*                                                                           */
/* liburl2cat is free software: you can redistribute it and/or modify it     */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation, either version 3 of the License, or             */
/* (at your option) any later version.                                       */
/*                                                                           */
/* liburl2cat is distributed in the hope that it will be useful, but         */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      */
/* See the GNU General Public License for more details.                      */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program.  If not, see <http://www.gnu.org/licenses/>.           */
/*                                                                           */
/*****************************************************************************/

#include "common.h"
#include "net.h"
#include "locale.h"

/*****************************************************************************/

static pthread_mutex_t server_mutex;

static char REQUEST_SLASH[] = "/";
static char REQUEST_COLON[] = ":";
static char REQUEST_END[] = "\r\n\r\n";
#define REQUEST_END_SIZE 4

/*request_get_part_  -> rgp_ */
/*REQUEST_GET_PART_  -> RGP_ */
static char* rgp_buffer = NULL;
static size_t rgp_buffer_size = 0;
static char* rgp_url_str = NULL;
static size_t rgp_url_str_size = 0;
static size_t rgp_buffer_begin_size = 0;
static char RGP_STR[] = "GET ";
#define RGP_STR_SIZE             (4 + 1)
#define RGP_URL_STR_SIZE_DEFAULT 512

static int malloc_request_get_str(char* path_str)
{
	size_t len = strlen(path_str);
	rgp_buffer_begin_size = len + RGP_STR_SIZE;
	rgp_url_str_size = RGP_URL_STR_SIZE_DEFAULT;
	size_t size = rgp_buffer_begin_size + rgp_url_str_size;
	rgp_buffer = malloc(size);
	if (rgp_buffer == NULL)
		return SERVER_REQUEST_GET_PATH_MALLOC_NULL;

	*rgp_buffer = 0;
	strcat(rgp_buffer, RGP_STR);
	strcat(rgp_buffer, path_str);
	strcat(rgp_buffer, REQUEST_SLASH);

	rgp_url_str = rgp_buffer + rgp_buffer_begin_size;
	return SUCCESS;
}

static int realloc_get_str(size_t size)
{
	size_t temp_size = rgp_url_str_size;
rgp_increment_size:
	temp_size <<= 1;
	if (temp_size < size)
		goto rgp_increment_size;
	char* temp_str = malloc(temp_size + rgp_buffer_begin_size);
	if (temp_str == NULL)
		return SERVER_REQUEST_GET_PATH_REALLOC_NULL;

	memmove(temp_str, rgp_buffer, rgp_buffer_begin_size);
	free(rgp_buffer);
	rgp_buffer = temp_str;
	rgp_url_str = rgp_buffer + rgp_buffer_begin_size;
	rgp_url_str_size = temp_size;
	return SUCCESS;
}

static void free_request_get_str(void)
{
	if (rgp_buffer != NULL) {
		free(rgp_buffer);
		rgp_buffer = NULL;
		rgp_buffer_size = 0;
	}
}

/*****************************************************************************/
/*request_host_part_ -> rhp*/
/*REQUEST_HOST_PART_ -> RHP*/
static char* rhp_buffer = NULL;
static size_t rhp_buffer_size = 0;
static char RHP_HTTP_STR[] = " HTTP/1.1\r\n"                            // 11
                             "Host: ";                                  // 6
                                                                        //"<host_str>:<port_str>"            // 0
static char RHP_USER_AGENT_STR[] = "\r\n"                               // 2
                                   "User-Agent: url2cat-00.000\r\n"     // 28
                                   "Accept: application/json\r\n"       // 26
                                   "Accept-Language: ";                 // 17
                                                                        //"locale
static char RHP_AUTHORIZATION_STR[] = "\r\n"                            // 2
                                      "Authorization: Basic ";          // 21
#define RHP_BUFFER_SIZE_DEFAULT (11 + 6 + 2 + 28 + 26 + 17 + 2 + 21)
static int malloc_request_host_str(char* host_str, char* port_str)
{
	size_t len_host_str = strlen(host_str);
	size_t len_port_str = 0;
	len_port_str = strlen(port_str);
	if (len_port_str != 0)
		++len_port_str;     // REQUEST_COLON
	size_t len_locale_str = strlen(LOCALE_NAME);
	rhp_buffer_size = RHP_BUFFER_SIZE_DEFAULT + len_host_str + len_port_str + len_locale_str;
	rhp_buffer = malloc(rhp_buffer_size + 1);
	if (rhp_buffer == NULL)
		return SERVER_REQUEST_HOST_MALLOC_NULL;
	*rhp_buffer = 0;
	strcat(rhp_buffer, RHP_HTTP_STR);
	strcat(rhp_buffer, host_str);
	if (len_port_str != 0) {
		strcat(rhp_buffer, REQUEST_COLON);
		strcat(rhp_buffer, port_str);
	}

	char version_major[3] = { 0 };
	snprintf(version_major, 3, "%02d", PROJECT_VERSION_MAJOR);
	RHP_USER_AGENT_STR[22] = version_major[0];
	RHP_USER_AGENT_STR[23] = version_major[1];
	char version_minor[4] = { 0 };
	snprintf(version_minor, 4, "%03d", PROJECT_VERSION_MINOR);
	RHP_USER_AGENT_STR[25] = version_minor[0];
	RHP_USER_AGENT_STR[26] = version_minor[1];
	RHP_USER_AGENT_STR[27] = version_minor[2];
	strcat(rhp_buffer, RHP_USER_AGENT_STR);
	strcat(rhp_buffer, LOCALE_NAME);
	strcat(rhp_buffer, RHP_AUTHORIZATION_STR);

	return SUCCESS;
}

static void free_request_host_str(void)
{
	if (rhp_buffer != NULL) {
		free(rhp_buffer);
		rhp_buffer = NULL;
		rhp_buffer_size = 0;
	}
}

/*****************************************************************************/
/*request_base64_part_ -> rbp*/
static char* rbp_buffer = NULL;
static size_t rbp_buffer_size = 0;
static int malloc_request_base64_str(char* user_str, char* password_str)
{
	size_t len_user_str = strlen(user_str);
	size_t len_password_str = strlen(password_str);
	char credentials_str[(len_user_str + len_password_str + 1 + 1)];
	*credentials_str = 0;
	strcat(credentials_str, user_str);
	strcat(credentials_str, REQUEST_COLON);
	strcat(credentials_str, password_str);
	int rc = encode_base64((uint8_t*)credentials_str, (len_user_str + len_password_str + 1), (uint8_t**)&rbp_buffer, &rbp_buffer_size);
	if (rc != SUCCESS)
		return rc;
	return SUCCESS;
}

static void free_request_base64_str(void)
{
	if (rbp_buffer != NULL) {
		free(rbp_buffer);
		rbp_buffer = NULL;
	}
}

/*****************************************************************************/
static char* request_buffer = NULL;

static int malloc_request_buffer()
{
	if (request_buffer != NULL)
		free(request_buffer);
	size_t size = rgp_buffer_begin_size + rgp_url_str_size + rhp_buffer_size + rbp_buffer_size + REQUEST_END_SIZE;
	request_buffer = malloc(size);
	if (request_buffer == NULL)
		return SERVER_REQUEST_BUFFER_MALLOC_NULL;
	return SUCCESS;
}

static void free_request_buffer(void)
{
	if (request_buffer != NULL) {
		free(request_buffer);
		request_buffer = NULL;
	}
}

/*****************************************************************************/
static char* answer_buffer = NULL;
static size_t answer_buffer_size = 0;

static int malloc_answer_buffer(size_t size)
{
	if (answer_buffer != NULL)
		free(answer_buffer);
	answer_buffer = malloc(size);
	if (answer_buffer == NULL)
		return SERVER_ANSWER_BUFFER_MALLOC_NULL;
	answer_buffer_size = size;
	return SUCCESS;
}

static void free_answer_buffer(void)
{
	if (answer_buffer != NULL) {
		free(answer_buffer);
		answer_buffer = NULL;
	}
}

/*****************************************************************************/

static int formation_get_str(char* url_str, size_t url_str_size)
{
	int rc;
	if (url_str_size > rgp_url_str_size) {
		rc = realloc_get_str(url_str_size);
		if (rc != SUCCESS)
			return rc;
	}

	memmove(rgp_url_str, url_str, url_str_size);
	rgp_buffer_size = rgp_buffer_begin_size + url_str_size;
	return SUCCESS;
}

static char ANSWER_NOT_FOUND_STR[] = "HTTP/1.1 404 Not Found";
#define ANSWER_NOT_FOUND_STR_SIZE 22
static char ANSWER_OK_STR[] = "HTTP/1.1 200 OK";
#define ANSWER_OK_STR_SIZE 15
static char CATEGORY_TYPE_STR[] = "\"category\": [";
#define CATEGORY_TYPE_STR_SIZE 13
/*                                                                           */
#define DECODE_SERVER 0
#if DECODE_SERVER
static char CATEGORY_TYPE_NAME_STR[] = "\"category_name\": [";
	#define CATEGORY_TYPE_NAME_STR_SIZE 18
#endif

#ifdef DEBUG
int parse_answer(char* answer_str, size_t answer_str_size, url2cat_category_s* p_category, int* p_number)
#else
static int parse_answer(char* answer_str, size_t answer_str_size, url2cat_category_s* p_category, int* p_number)
#endif
{
	int rc;
	/*            HTTP           */
	rc = strncmp(answer_str, ANSWER_NOT_FOUND_STR, ANSWER_NOT_FOUND_STR_SIZE);
	if (rc == 0) {
		*p_number = 0;
		return SERVER_URL_NOT_FOUND;
	}
	rc = strncmp(answer_str, ANSWER_OK_STR, ANSWER_OK_STR_SIZE);
	if (rc != 0) {
		*p_number = 0;
		return SERVER_ANSWER_NOT_OK;
	}

	char* str_category_type = NULL;
	str_category_type = strstr(answer_str, CATEGORY_TYPE_STR);
	if (str_category_type == NULL) {
		*p_number = 0;
		return SERVER_ANSWER_NOT_CATEGORY_TYPE;
	}
	str_category_type += CATEGORY_TYPE_STR_SIZE;
	char* str_category_type_end = strchr(str_category_type, ']');
	if (str_category_type_end == NULL) {
		*p_number = 0;
		return SERVER_ANSWER_NOT_CATEGORY_TYPE_BRACKET;
	}
	*str_category_type_end = 0;

#if DECODE_SERVER
	char* str_category_type_name = NULL;
	str_category_type_name = strstr(answer_str, CATEGORY_TYPE_NAME_STR);
	if (str_category_type_name == NULL) {
		*p_number = 0;
		return SERVER_ANSWER_NOT_CATEGORY_TYPE_NAME;
	}
	str_category_type_name += CATEGORY_TYPE_NAME_STR_SIZE;
	char* str_category_type_name_end = strchr(str_category_type_name, ']');
	if (str_category_type_name_end == NULL) {
		*p_number = 0;
		return SERVER_ANSWER_NOT_CATEGORY_TYPE_NAME_BRACKET;
	}
	*str_category_type_name_end = 0;
#endif
	int number;
	for (number = 0; number < *p_number; ++number, ++p_category) {
		char* str_temp = strchr(str_category_type, ',');
		if (str_temp != NULL) {
			*str_temp = 0;
			++str_temp;
		}
		else {
			str_temp = str_category_type_end;
		}
		p_category->type = strtol(str_category_type, NULL, 10);
		if (p_category->type == 0) {
			*p_number = number;
			return SERVER_ANSWER_NOT_CATEGORY_TYPE_INT;
		}

		str_category_type = str_temp;
#if DECODE_SERVER
		str_temp = strchr(str_category_type_name, ',');
		if (str_temp != NULL) {
			--str_temp; /*                 */
			*str_temp = 0;
			str_temp += 2;
		}
		else {
			--str_category_type_name_end;
			*str_category_type_name_end = 0;
		}
		strncpy(p_category->type_name, (str_category_type_name + 1), STR_SIZE_TYPE_NAME_CATEGORY);
		str_category_type_name = str_temp;
#else
		const char* name = get_category_name(p_category->type);
		if (name != NULL)
			strncpy(p_category->type_name, name, CATEGORY_TYPE_NAME_STR_SIZE);
		else
			*p_category->type_name = 0;
#endif
		if (str_category_type == str_category_type_end) {
			*p_number = number;
			break;
		}
	}
	if (*p_number > number)
		*p_number = number;

	return SUCCESS;
}

static net_option_s net_option = { 0 };

int get_category_server(char* url_str, size_t url_str_size, url2cat_category_s* p_category, int* p_number)
{
	int rc;
	net_connect_s net_connect;

	rc = connect_net(&net_option, &net_connect);
	if (rc != SUCCESS)
		goto get_category_server_exit_error;

	pthread_mutex_lock(&server_mutex);

	rc = formation_get_str(url_str, url_str_size);
	if (rc != SUCCESS) {
		pthread_mutex_unlock(&server_mutex);
		goto get_category_server_exit_error;
	}

	rc = malloc_request_buffer();
	if (rc != SUCCESS) {
		pthread_mutex_unlock(&server_mutex);
		goto get_category_server_exit_error;
	}
	rc = malloc_answer_buffer(ANSWER_BUFFER_SIZE_DEFAULT);
	if (rc != SUCCESS) {
		pthread_mutex_unlock(&server_mutex);
		goto get_category_server_exit_error;
	}

	char* request_str = request_buffer;
	size_t request_str_size = 0;
	memmove(request_str, rgp_buffer, rgp_buffer_size);
	request_str += rgp_buffer_size;
	request_str_size += rgp_buffer_size;
	pthread_mutex_unlock(&server_mutex);

	memmove(request_str, rhp_buffer, rhp_buffer_size);
	request_str += rhp_buffer_size;
	request_str_size += rhp_buffer_size;

	memmove(request_str, rbp_buffer, rbp_buffer_size);
	request_str += rbp_buffer_size;
	request_str_size += rbp_buffer_size;

	memmove(request_str, REQUEST_END, REQUEST_END_SIZE);
	request_str_size += REQUEST_END_SIZE;

	rc = send_net(&net_connect, request_buffer, request_str_size);
	if (rc != request_str_size) {
		rc = SERVER_CATEGORY_HOST_WRITE;
		goto get_category_server_exit_error;
	}

	rc = recv_net(&net_connect, answer_buffer, answer_buffer_size);
	if (rc <= 0) {
		rc = SERVER_CATEGORY_HOST_READ;
		goto get_category_server_exit_error;
	}

	rc = parse_answer(answer_buffer, rc, p_category, p_number);

get_category_server_exit_error:
	if (rc != SUCCESS)
		*p_number = 0;
	free_request_buffer();
	free_answer_buffer();
	disconnect_net(&net_connect);

	return rc;
}

int init_server(char* scheme_str, char* host_str, char* port_str, char* path_str, char* user_str, char* password_str)
{
	int rc = malloc_request_get_str(path_str);
	if (rc != SUCCESS)
		return rc;

	rc = malloc_request_host_str(host_str, port_str);
	if (rc != SUCCESS) {
		free_request_get_str();
		return rc;
	}

	rc = malloc_request_base64_str(user_str, password_str);
	if (rc != SUCCESS) {
		free_request_get_str();
		free_request_host_str();
		return rc;
	}

	rc = init_net(&net_option, scheme_str, host_str, port_str);
	if (rc != SUCCESS) {
		free_request_get_str();
		free_request_host_str();
		free_request_base64_str();
		free_request_buffer();
		free_answer_buffer();
	}
	pthread_mutex_init(&server_mutex, NULL);
	return SUCCESS;
}

int deinit_server(void)
{
	free_request_get_str();
	free_request_host_str();
	free_request_base64_str();
	deinit_net(&net_option);
	pthread_mutex_destroy(&server_mutex);
	return SUCCESS;
}

/*****************************************************************************/
