/*****************************************************************************/
/*                                                                           */
/* liburl2cat : Domain Categorization Library                                */
/*                                                                           */
/*  Copyright (C) 2020 Kuzmin Yaroslav <kuzmin.yaroslav@gmail.com>           */
/*  Copyright (C) 2022 Kuzmin Yaroslav <kuzmin.yaroslav@gmail.com>           */
/*                                                                           */
/* liburl2cat is free software: you can redistribute it and/or modify it     */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation, either version 3 of the License, or             */
/* (at your option) any later version.                                       */
/*                                                                           */
/* liburl2cat is distributed in the hope that it will be useful, but         */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      */
/* See the GNU General Public License for more details.                      */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program.  If not, see <http://www.gnu.org/licenses/>.           */
/*                                                                           */
/*****************************************************************************/

#include "version_git.h"
#include "common.h"
#include "canonicalize.h"
#include "cache.h"
#include "database.h"
#include "server.h"
#include "locale.h"
#include "recategory.h"
#include "net.h"

/*****************************************************************************/
#define PROJECT_STR_SIZE 32
static char PROJECT[PROJECT_STR_SIZE] = { 0 };
const char* url2cat_version(void)
{
	if (PROJECT[0] == 0)
		snprintf(PROJECT, PROJECT_STR_SIZE, "%s %02d.%03d-%x", PROJECT_NAME, PROJECT_VERSION_MAJOR, PROJECT_VERSION_MINOR, PROJECT_VERSION_GIT);
	return PROJECT;
}

/*****************************************************************************/
struct option
{
	char* key;
	char* value;
};
typedef struct option option_s;

#define DEFAULT_NUMBER_OPTIONS 16

static int realloc_options(option_s** options, int* number)
{
	if (options == NULL)
		return READ_INI_OPTIONS;

	if (*options == NULL) {
		*options = calloc(DEFAULT_NUMBER_OPTIONS, sizeof(option_s));
		if (*options == NULL)
			return READ_INI_CALLOC_OPTIONS;
		*number = DEFAULT_NUMBER_OPTIONS;
	}
	else {
		int new_number = *number + DEFAULT_NUMBER_OPTIONS;
		option_s* new_options = calloc(new_number, sizeof(option_s));
		if (new_options == NULL)
			return READ_INI_RECALLOC_OPTIONS;
		memmove(new_options, *options, (*number * sizeof(option_s)));
		free(*options);
		*options = new_options;
		*number = new_number;
	}
	return SUCCESS;
}

static int free_options(option_s* options)
{
	if (options == NULL)
		return SUCCESS;
	free(options);
	return SUCCESS;
}

static option_s* begin_options = NULL;
static int number_options = 0;
static int max_number_options = 0;

static int fill_options(int fd, size_t size)
{
	char* buff;
	FILE* stream;
	char* str;

	stream = fdopen(fd, "r");
	if (stream == NULL)
		return READ_INI_NOT_STREAM;
	buff = malloc(size);
	if (buff == NULL)
		return READ_INI_MALLOC_BUFF;

	for (;;) {
		size_t len;
		char* key;
		char* value;

		str = fgets(buff, size, stream);
		if (str == NULL)
			break;
		if (*str < 'a')
			continue;
		len = strlen(str);

		key = malloc(len);
		if (key == NULL)
			break;
		strncpy(key, str, len);
		value = strchr(key, '=');
		if (value == NULL) {
			free(key);
			continue;
		}
		*value = 0;
		++value;
		str = strchr(value, '\n');
		if (str != NULL)
			*str = 0;
		if (number_options == max_number_options)
			realloc_options(&begin_options, &max_number_options);

		(begin_options + number_options)->key = key;
		(begin_options + number_options)->value = value;
		++number_options;
	}

	free(buff);
	return SUCCESS;
}

static int del_options(void)
{
	option_s* curr = begin_options;

	for (int i = 0; i < number_options; ++i) {
		if (curr->key != NULL)
			free(curr->key);
		++curr;
	}
	free_options(begin_options);
	number_options = 0;
	max_number_options = 0;
	begin_options = NULL;
	return SUCCESS;
}

static int get_option(char* option, char** value, size_t* size)
{
	int rc;
	option_s* curr = begin_options;
	*value = NULL;
	*size = 0;
	for (int i = 0; i < number_options; ++i) {
		rc = strcmp(curr->key, option);
		if (rc != 0) {
			++curr;
			continue;
		}
		*value = curr->value;
		*size = strlen(curr->value);
		break;
	}
	return SUCCESS;
}

int url2cat_read_ini(char* file_name, url2cat_setting_s* setting)
{
	int rc;
	int fd;
	size_t len;
	char* value;

	fd = open(file_name, O_RDONLY);
	if (fd == -1)
		return READ_INI_NOT_OPEN_FILE;
	len = lseek(fd, 0, SEEK_END);
	++len;
	lseek(fd, 0, SEEK_SET);
	rc = fill_options(fd, len);
	if (rc != SUCCESS) {
		del_options();
		close(fd);
		return rc;
	}

	get_option("cache_size", &value, &len);
	if (len != 0)
		setting->cache_size = strtol(value, NULL, 10);
#ifdef HAVE_DATABASE
	get_option("db_name", &value, &len);
	if (len != 0) {
		if (len > DATABASE_NAME_STR_SIZE)
			len = DATABASE_NAME_STR_SIZE;
		strncpy(setting->db_name, value, len);
	}

	get_option("db_update_host", &value, &len);
	if (len != 0) {
		if (len > NET_HOST_STR_SIZE)
			len = NET_HOST_STR_SIZE;
		strncpy(setting->db_update_host, value, len);
	}

	get_option("db_update_path", &value, &len);
	if (len != 0) {
		if (len > NET_PATH_STR_SIZE)
			len = NET_PATH_STR_SIZE;
		strncpy(setting->db_update_path, value, len);
	}

	get_option("db_download_path", &value, &len);
	if (len != 0) {
		if (len > NET_PATH_STR_SIZE)
			len = NET_PATH_STR_SIZE;
		strncpy(setting->db_download_path, value, len);
	}

	get_option("db_update_user", &value, &len);
	if (len != 0) {
		if (len > NET_USER_STR_SIZE)
			len = NET_USER_STR_SIZE;
		strncpy(setting->db_update_user, value, len);
	}

	get_option("db_update_password", &value, &len);
	if (len != 0) {
		if (len > NET_PASSWORD_STR_SIZE)
			len = NET_PASSWORD_STR_SIZE;
		strncpy(setting->db_update_password, value, len);
	}

	get_option("db_user_version", &value, &len);
	if (len != 0) {
		if (len > DATABASE_USER_VERSION_SIZE)
			len = DATABASE_USER_VERSION_SIZE;
		strncpy(setting->db_user_version, value, len);
	}
#endif

#ifdef HAVE_SERVER
	get_option("server_host", &value, &len);
	if (len != 0) {
		if (len > NET_HOST_STR_SIZE)
			len = NET_HOST_STR_SIZE;
		strncpy(setting->server_host, value, len);
	}

	get_option("server_path", &value, &len);
	if (len != 0) {
		if (len > NET_PATH_STR_SIZE)
			len = NET_PATH_STR_SIZE;
		strncpy(setting->server_path, value, len);
	}

	get_option("server_user", &value, &len);
	if (len != 0) {
		if (len > NET_USER_STR_SIZE)
			len = NET_USER_STR_SIZE;
		strncpy(setting->server_user, value, len);
	}

	get_option("server_password", &value, &len);
	if (len != 0) {
		if (len > NET_PASSWORD_STR_SIZE)
			len = NET_PASSWORD_STR_SIZE;
		strncpy(setting->server_password, value, len);
	}
#endif

	get_option("recat_host", &value, &len);
	if (len != 0) {
		if (len > NET_HOST_STR_SIZE)
			len = NET_HOST_STR_SIZE;
		strncpy(setting->recat_host, value, len);
	}

	get_option("recat_path", &value, &len);
	if (len != 0) {
		if (len > NET_PATH_STR_SIZE)
			len = NET_PATH_STR_SIZE;
		strncpy(setting->recat_path, value, len);
	}

	get_option("recat_user", &value, &len);
	if (len != 0) {
		if (len > NET_USER_STR_SIZE)
			len = NET_USER_STR_SIZE;
		strncpy(setting->recat_user, value, len);
	}

	get_option("recat_password", &value, &len);
	if (len != 0) {
		if (len > NET_PASSWORD_STR_SIZE)
			len = NET_PASSWORD_STR_SIZE;
		strncpy(setting->recat_password, value, len);
	}

	del_options();
	close(fd);
	return SUCCESS;
}

/*****************************************************************************/
static char DEFAULT_SCHEME[] = "https";
static char DEFAULT_PORT[] = "";
static int flag_init = NO;

int url2cat_init(url2cat_setting_s* setting)
{
	int rc;

	if (flag_init == YES)
		return INIT_REPEATED;

#if defined(HAVE_DATABASE) && !defined(HAVE_SERVER)
	setting->cache_size = 0;
#endif
	rc = init_cache(setting->cache_size);
	if (rc != SUCCESS)
		return rc;

#ifdef HAVE_DATABASE
	if (*setting->db_name == 0)
		return INIT_DB_NAME_EMPTY;
	if (*setting->db_update_host == 0)
		return INIT_DB_UPDATE_HOST_EMPTY;
	if (*setting->db_update_path == 0)
		return INIT_DB_UPDATE_PATH_EMPTY;
	if (*setting->db_download_path == 0)
		return INIT_DB_DOWNLOAD_PATH_EMPTY;
	if (*setting->db_update_user == 0)
		return INIT_DB_UPDATE_USER_EMPTY;
	if (*setting->db_update_password == 0)
		return INIT_DB_UPDATE_PASSWORD_EMPTY;

	setting->db_name[DATABASE_NAME_STR_SIZE - 1] = 0;
	setting->db_update_host[NET_HOST_STR_SIZE - 1] = 0;
	setting->db_update_path[NET_PATH_STR_SIZE - 1] = 0;
	setting->db_download_path[NET_PATH_STR_SIZE - 1] = 0;
	setting->db_update_user[NET_USER_STR_SIZE - 1] = 0;
	setting->db_update_password[NET_PASSWORD_STR_SIZE - 1] = 0;
	setting->db_user_version[DATABASE_USER_VERSION_SIZE - 1] = 0;

	rc = init_database(setting->db_name, DEFAULT_SCHEME, setting->db_update_host, DEFAULT_PORT, setting->db_update_path, setting->db_download_path, setting->db_update_user, setting->db_update_password);
	const char* db_user_version = get_user_version_database();
	memmove(setting->db_user_version, db_user_version, DATABASE_USER_VERSION_SIZE);
	if (rc != SUCCESS) {
		deinit_cache();
		return rc;
	}
#endif

#ifdef HAVE_SERVER
	if (*setting->server_host == 0)
		return INIT_SERVER_HOST_EMPTY;
	if (*setting->server_path == 0)
		return INIT_SERVER_PATH_EMPTY;
	if (*setting->server_user == 0)
		return INIT_SERVER_USER_EMPTY;
	if (*setting->server_password == 0)
		return INIT_SERVER_PASSWORD_EMPTY;

	setting->server_host[NET_HOST_STR_SIZE - 1] = 0;
	setting->server_path[NET_PATH_STR_SIZE - 1] = 0;
	setting->server_user[NET_USER_STR_SIZE - 1] = 0;
	setting->server_password[NET_PASSWORD_STR_SIZE - 1] = 0;

	rc = init_server(DEFAULT_SCHEME, setting->server_host, DEFAULT_PORT, setting->server_path, setting->server_user, setting->server_password);
	if (rc != SUCCESS) {
		deinit_cache();
	#ifdef HAVE_DATABASE
		deinit_database();
	#endif
		return rc;
	}
#endif

	if (*setting->recat_host == 0)
		return INIT_RECAT_HOST_EMPTY;
	if (*setting->recat_path == 0)
		return INIT_RECAT_PATH_EMPTY;
	if (*setting->recat_user == 0)
		return INIT_RECAT_USER_EMPTY;
	if (*setting->recat_password == 0)
		return INIT_RECAT_PASSWORD_EMPTY;

	setting->recat_host[NET_HOST_STR_SIZE - 1] = 0;
	setting->recat_path[NET_PATH_STR_SIZE - 1] = 0;
	setting->recat_user[NET_USER_STR_SIZE - 1] = 0;
	setting->recat_password[NET_PASSWORD_STR_SIZE - 1] = 0;

	rc = init_recategory(DEFAULT_SCHEME, setting->recat_host, DEFAULT_PORT, setting->recat_path, setting->recat_user, setting->recat_password);
	if (rc != SUCCESS) {
		deinit_cache();
#ifdef HAVE_DATABASE
		deinit_database();
#endif
#ifdef HAVE_SERVER
		deinit_server();
#endif
		return rc;
	}

	flag_init = YES;
	return rc;
}

int url2cat_deinit(void)
{
	if (flag_init == NO)
		return SUCCESS;
	deinit_canonicalize_url();
	deinit_cache();
	deinit_recategory();
#ifdef HAVE_DATABASE
	deinit_database();
#endif
#ifdef HAVE_SERVER
	deinit_server();
#endif
	free_net();
	flag_init = NO;
	return SUCCESS;
}

struct url_check
{
	char* begin;
	int size;
	char* begin_path;
	int size_path;
	int size_path_end;
	int size_domain;
	int size_domain_root;
};
typedef struct url_check url_check_s;

static int init_url_check(url_check_s* url_check, char* url, int url_size)
{
	char* domain_end;

	url_check->begin = url;
	url_check->begin_path = memchr(url, '/', url_size);
	if (url_check->begin_path == NULL)
		return URL_INCORRECT;

	url_check->size_path = url_size - (url_check->begin_path - url_check->begin);
	url_check->size_domain = url_size - url_check->size_path;
	url_check->size_domain_root = 0;
	domain_end = memrchr(url, '.', url_check->size_domain);
	if (domain_end == NULL)
		return URL_INCORRECT;
	else {
		url_check->size_domain_root = (url_check->begin_path - domain_end);
		url_check->size_domain = (domain_end - url);
	}

	url_check->size = url_check->size_domain + url_check->size_domain_root + url_check->size_path;
	url_check->size_path_end = url_check->size_path;
	return SUCCESS;
}

static int get_url_check(url_check_s* url_check)
{
	char* domain_end;
	char* path_end = memrchr(url_check->begin_path, '/', url_check->size_path_end);
	if (url_check->begin_path != path_end) {
		url_check->size_path_end = path_end - url_check->begin_path;
		url_check->size = url_check->size_domain + url_check->size_domain_root + url_check->size_path_end;
	}
	else {
		if (url_check->size_path_end != 1) {
			url_check->size_path_end = 1;
			url_check->size = url_check->size_domain + url_check->size_domain_root + url_check->size_path_end;
			return SUCCESS;
		}

		domain_end = memchr(url_check->begin, '.', url_check->size_domain);
		if (domain_end == NULL)
			return URL_END_CHECK;
		++domain_end;
		url_check->size_domain -= (domain_end - url_check->begin);
		url_check->begin = domain_end;
		url_check->size = url_check->size_domain + url_check->size_domain_root + url_check->size_path;
		url_check->size_path_end = url_check->size_path;
	}
	return SUCCESS;
}

int url2cat_category(char* url, size_t url_size, url2cat_category_s* p_category, int* p_number)
{
	int rc;
	char* url_canonical = NULL;
	size_t url_canonical_size = 0;
	url_check_s url_check;

	if ((p_category == NULL) || (p_number == NULL) || (*p_number == 0))
		return CATEGORY_ARRAY_NULL;

	rc = get_canonicalize_url(url, url_size, &url_canonical, &url_canonical_size);
	if (rc != SUCCESS)
		goto url2cat_category_exit;

	rc = init_url_check(&url_check, url_canonical, url_canonical_size);
	if (rc != SUCCESS)
		goto url2cat_category_exit;

	for (;;) {
		rc = get_category_cache(url_check.begin, url_check.size, p_category, p_number);
		if (rc == SUCCESS)
			goto url2cat_category_exit;

#ifdef HAVE_DATABASE
		rc = get_category_database(url_check.begin, url_check.size, p_category, p_number);
		if ((rc == SUCCESS) || (rc != DB_URL_NOT_FOUND))
			goto url2cat_category_exit;
#endif

#ifdef HAVE_SERVER
		rc = get_category_server(url_check.begin, url_check.size, p_category, p_number);
		if ((rc == SUCCESS) || (rc != SERVER_URL_NOT_FOUND))
			goto url2cat_category_exit;
#endif

		rc = get_url_check(&url_check);
		if (rc == URL_END_CHECK)
			break;
	}

	p_category->type = unknown_sites;
	const char* type_name = get_category_name(unknown_sites);
	if (type_name != NULL)
		strncpy(p_category->type_name, type_name, CATEGORY_TYPE_NAME_STR_SIZE);
	else
		*p_category->type_name = 0;
	*p_number = 1;
	rc = SUCCESS;

url2cat_category_exit:
	if (rc != SUCCESS)
		*p_number = 0;
	add_category_cache(url_canonical, url_canonical_size, p_category, *p_number);
	if (url_canonical != NULL)
		free(url_canonical);
	return rc;
}

#ifdef HAVE_DATABASE
int url2cat_update_database(url2cat_setting_s* setting)
{
	int rc = update_database();
	if (rc != SUCCESS)
		return rc;

	const char* db_user_version = get_user_version_database();
	memmove(setting->db_user_version, db_user_version, DATABASE_USER_VERSION_SIZE);

	return SUCCESS;
}
#endif

int url2cat_recategory(char* url, size_t url_size, char* category_name, size_t category_name_size)
{
#if 0
	char* url_canonical = NULL;
	size_t url_canonical_size = 0;
	int rc = get_canonicalize_url(url, url_size
	                              , &url_canonical, &url_canonical_size);
	if (rc != SUCCESS)
		return rc;
#endif

	return send_recategory(url, url_size, category_name, category_name_size);
}
/*****************************************************************************/
